"""
Pipeline to create polar diagrams from raw data.
"""


import warnings
from abc import ABC, abstractmethod
from datetime import datetime as dt
from typing import NamedTuple

import numpy as np

import hrosailing.pipelinecomponents as pc
import hrosailing.polardiagram as pol
from hrosailing.pipelinecomponents.modelfunctions import (
    ws_s_wa_gauss_and_square,
)

from .extensions import (
    CurveExtension,
    PipelineExtension,
    PointcloudExtension,
    TableExtension,
)


class Statistics(NamedTuple):
    """
    Organizes the statistics returned by different `Pipelinecomponents`.
    The attributes correspond to the parameters of `PolarPipeline.__init__`
    and each contains dictionaries with relevant statistics.

    See also
    ----------
    `PolarPipeline`
    """

    data_handler: dict
    imputator: dict
    smoother: dict
    pre_expanding_weigher: dict
    pre_expanding_filter: dict
    expander: dict
    pre_influence_weigher: dict
    pre_influence_filter: dict
    influence_model: dict
    post_weigher: dict
    post_filter: dict
    injector: dict
    extension: dict
    quality_assurance: dict


_EMPTY_STATISTIC = Statistics(
    {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}
)


class PipelineOutput(NamedTuple):
    """
    Organizes the output of a `PolarPipeline` call.

    Attributes
    ----------
    polardiagram: PolarDiagram
        The resulting polar diagram.

    training_statistics: Statistics
        Relevant statistics obtained in the preprocessing and processing of the training data.

    test_statistics: Statistics
        Relevant statistics obtained in the preprocessing of the test data.
        The attributes `extension` and `quality_assurance` only contain an empty dictionary.
    """

    polardiagram: pol.PolarDiagram
    training_statistics: Statistics
    test_statistics: Statistics


class PolarPipeline:
    """A Pipeline class to create polar diagrams from raw data.

    Other Parameters
    ---------------------------
    data_handler : DataHandler or list of DataHandler, optional
        Handlers that are responsible to extract actual data from the input.
        If only one handler is given, this handler will be used for all given inputs,
        otherwise the handlers will be used one after another for each data input including the training data.

        Determines the type and format of input the pipeline should accept.

    imputator : Imputator, optional
        Determines the method which will be used to produce data without
        `None` entries.

        Defaults to `RemoveOnlyImputator()`.

    smoother: Smoother, optional
        Determines the method which will be used to smoothen out the rounding
        following from low measurement precision.

        Defaults to `LazySmoother()`.

    pre_expander_weigher : Weigher, optional
        Determines the method with which the points will be weighted before application of the expander.

        Defaults to `CylindricMeanWeigher()`

    pre_expander_filter : Filter, optional
        Determines the method with which the points will be filtered with before application of the expander.

        Defaults to `QuantileFilter()`

    expander: Expander, optional
        Determines the method which will be used to expand the data by several more data fields.
        For example weather data from a weather model.

    pre_influence_weigher, pre_weigher : Weigher, optional
        Determines the method with which the points will be weighted before
        application of the influence model.

        Defaults to `CylindricMeanWeigher()`.

    pre_influence_filter, pre_filter : Filter, optional
        Determines the methods which the points will be filtered with before application of the influence model.

        Defaults to `QuantileFilter()`.

    influence_model : InfluenceModel, optional
        Determines the influence model which is applied and fitted to the data.

        Defaults to 'IdentityInfluenceModel()'.

    post_weigher, weigher : Weigher, optional
        Determines the method with which the points will be weighted after
        application of the influence model.

        Defaults to `CylindricMeanWeigher()`.

    post_filter, filter : Filter, optional
        Determines the methods with which the points will be filtered
        after the application of the influence model,
        if `post_filtering` in `__call__` method.

        Defaults to `QuantileFilter()`.

    injector : Injector, optional
        Determines the method used to add additional artificial data points to the
        data.

        Defaults to `ZeroInjector(500)`.

    extension: PipelineExtension
        Extension that is called in the pipeline, after all preprocessing
        is done, to generate a polar diagram from the processed data.

        Determines the subclass of `PolarDiagram`, that the pipeline will
        produce.

        Defaults to `TableExtension()`.

    quality_assurance : QualityAssurance, optional
        Determines the method which is used to measure the quality of the
        resulting polar diagram using preprocessed test_data.

        Defaults to `MinimalQualityAssurance()`.
    """

    def __init__(self, **custom_components):
        defaults = [
            pc.NMEAFileHandler(),
            pc.RemoveOnlyImputator(),
            pc.LazySmoother(),
            pc.CylindricMeanWeigher(),
            pc.QuantileFilter(),
            pc.LazyExpander(),
            pc.CylindricMeanWeigher(),
            pc.QuantileFilter(),
            pc.IdentityInfluenceModel(),
            pc.CylindricMeanWeigher(),
            pc.QuantileFilter(),
            pc.ZeroInjector(500),
            TableExtension(),
            pc.MinimalQualityAssurance(),
        ]

        keys = [
            "data_handler",
            "imputator",
            "smoother",
            "pre_expander_weigher",
            "pre_expander_filter",
            "expander",
            "pre_influence_weigher",
            "pre_influence_filter",
            "influence_model",
            "post_weigher",
            "post_filter",
            "injector",
            "extension",
            "quality_assurance",
        ]

        aliases = {
            "filter": "post_filter",
            "weigher": "post_weigher",
            "pre_filter": "pre_influence_filter",
            "pre_weigher": "pre_influence_weigher",
        }

        self._set_with_default(custom_components, keys, defaults, aliases)

    def __call__(
        self, training_data, test_data=None, apparent_wind=False, **enabling
    ):
        """
        Parameters
        ----------
        training_data : list of data compatible with `self.data_handler`
            Data from which to create the polar diagram.

            The input should be compatible with the DataHandler instances
            given in initialization of the pipeline instance.
            Also, the input should be suitable to be interpreted as chronologically
            ordered time series by the before-mentioned data handler.

        test_data: list of data compatible with `self.data_handler` or `None`
            Data which is preprocessed and then used to check the quality of
            the resulting polar diagram.

            The input should be compatible with the DataHandler instances
            given in initialization of the pipeline instance.
            If `None` no quality check is performed.

            Default to `None`.

        apparent_wind : bool, optional
            Specifies if wind data is given in apparent wind.

            If `True`, wind will be converted to true wind.

            Defaults to `False`.

        Other Parameters
        ---------------------------

        pre_expander_weighing : bool, optional
            Specifies, whether the pre_influence_weigher should be applied before application of the
            expander.
            Otherwise, each point will be assigned the weight 1.

            Defaults to `True`.

        pre_expander_filtering : bool, optional
            Specifies, whether the pre_influence_filter should be applied before application of the expander.

            Defaults to `True`.


        pre_influence_weighing, pre_weighing : bool, optional
            Specifies, whether the pre_influence_weigher should be applied before application of the
            influence model.
            Otherwise, each point will be assigned the weight 1.

            Defaults to `True`.

        pre_influence_filtering, pre_filtering : bool, optional
            Specifies, whether the pre_influence_filter should be applied before application of the influence model.

            Defaults to `True`.

        smoothing : bool, optional
            Specifies, if measurement errors of the time series should be
            smoothened after pre_filtering.

        post_weighing, weighing : bool, optional
            Specifies, if points should be weighed after application of the
            influence model.
            Otherwise, each point will be assigned the weight 1.

            Defaults to `True`.

        post_filtering, filtering : bool, optional
            Specifies, if points should be filtered after post_weighing.

            Defaults to `True`.

        injecting : bool, optional
            Specifies, if artificial points should be added to the data.

            Defaults to `True`.

        testing : bool, optional
            Specifies, if the resulting polar diagram should be tested against
            test data.

        Returns
        -------
        out : PipelineOutput
        """
        start_time = dt.now()
        keys = [
            "smoothing",
            "pre_expander_weighing",
            "pre_expander_filtering",
            "pre_influence_weighing",
            "pre_influence_filtering",
            "post_weighing",
            "post_filtering",
            "injecting",
            "testing",
        ]
        aliases = {
            "pre_weighing": "pre_influence_weighing",
            "pre_filtering": "pre_influence_filtering",
            "weighing": "post_weighing",
            "filtering": "post_filtering",
        }
        defaults = [True] * len(keys)

        self._set_with_default(enabling, keys, defaults, aliases)

        if test_data is None:
            self.testing = False

        preproc_training_data, pp_training_statistics = self._preprocess(
            training_data,
            True,
        )

        if is_empty_data(preproc_training_data):
            raise RuntimeError(
                "Empty data after preprocessing. Try to use weaker filters."
            )

        if self.injecting:
            pts_to_inject, injector_statistics = _collect(
                self.injector, self.injector.inject, preproc_training_data
            )
        else:
            pts_to_inject, injector_statistics = (
                pc.WeightedPoints(np.empty((0, 3)), np.empty(0)),
                {},
            )

        preproc_training_data.extend(pts_to_inject)

        polar_diagram, extension_statistics = _collect(
            self.extension, self.extension.process, preproc_training_data
        )

        if self.testing:
            preproc_test_data, test_statistics = self._preprocess(
                test_data,
                False,
            )
            quality_assurance_statistics = self.quality_assurance.check(
                polar_diagram, preproc_test_data.data
            )
        else:
            test_statistics = _EMPTY_STATISTIC
            quality_assurance_statistics = {}

        quality_assurance_statistics["execution time"] = (
            dt.now() - start_time
        ).total_seconds()

        training_statistics = Statistics(
            pp_training_statistics.data_handler,
            pp_training_statistics.imputator,
            pp_training_statistics.smoother,
            pp_training_statistics.pre_expanding_weigher,
            pp_training_statistics.pre_influence_filter,
            pp_training_statistics.expander,
            pp_training_statistics.pre_influence_weigher,
            pp_training_statistics.pre_influence_filter,
            pp_training_statistics.influence_model,
            pp_training_statistics.post_weigher,
            pp_training_statistics.post_filter,
            injector_statistics,
            extension_statistics,
            quality_assurance_statistics,
        )

        return PipelineOutput(
            polar_diagram, training_statistics, test_statistics
        )

    def _preprocess(self, data, influence_fitting):
        data, handler_statistics = self._handle_data(data)

        data, imputator_statistics = self._map(
            _collector_fun(self.imputator, self.imputator.impute), data
        )

        if self.smoothing:
            data, smooth_statistics = self._map(
                _collector_fun(self.smoother, self.smoother.smooth),
                data,
            )
        else:
            smooth_statistics = {}

        (
            data,
            pre_exp_weigher_statistics,
            pre_exp_filter_statistics,
        ) = self._map(
            lambda x: _weigh_and_filter(
                x,
                self.pre_expander_weigher,
                self.pre_expander_filter,
                self.pre_expander_weighing,
                self.pre_expander_filtering,
            ),
            data,
        )

        data = [weighted_point.data for weighted_point in data]

        data, expanded_statistics = self._map(
            _collector_fun(self.expander, self.expander.expand),
            data,
        )

        (
            data,
            pre_weigher_statistics,
            pre_filter_statistics,
        ) = self._map(
            lambda data: _weigh_and_filter(
                data,
                self.pre_influence_weigher,
                self.pre_influence_filter,
                self.pre_influence_weighing,
                self.pre_influence_filtering,
            ),
            data,
        )

        data = pc.data.Data.concatenate([wp.data for wp in data])

        if influence_fitting:
            self.influence_model.fit(data)
            influence_fit_statistics = (
                self.influence_model.get_latest_statistics()
            )
        else:
            influence_fit_statistics = {}

        data, influence_statistics = _collect(
            self.influence_model,
            self.influence_model.remove_influence,
            data,
        )

        influence_statistics.update(influence_fit_statistics)

        (
            data,
            post_weigher_statistics,
            post_filter_statistics,
        ) = _weigh_and_filter(
            data,
            self.post_weigher,
            self.post_filter,
            self.post_weighing,
            self.post_filtering,
        )

        statistics = Statistics(
            handler_statistics,
            imputator_statistics,
            smooth_statistics,
            pre_exp_weigher_statistics,
            pre_exp_filter_statistics,
            expanded_statistics,
            pre_weigher_statistics,
            pre_filter_statistics,
            influence_statistics,
            post_weigher_statistics,
            post_filter_statistics,
            {},
            {},
            {},
        )

        return data, statistics

    def _set_with_default(self, dict_, keys, defaults, aliases):
        for key in list(dict_.keys()):
            if key in aliases:
                dict_[aliases[key]] = dict_[key]
        for key, default in zip(keys, defaults):
            setattr(self, key, self._switch_default(dict_, key, default))

    def _switch_default(self, dict_, key, default):
        return dict_[key] if key in dict_ else default

    def _handle_data(self, data):
        if isinstance(self.data_handler, pc.DataHandler):
            handler_output = [
                _collect(self.data_handler, self.data_handler.handle, field)
                for field in data
            ]
        else:
            handler_output = []
            for field in data:
                handler = next(self.data_handler)
                handler_output.append(_collect(handler, handler.handle, field))

        return self._list_statistics(handler_output)

    def _map(self, method, data):
        output = [method(field) for field in data]
        return self._list_statistics(output)

    @staticmethod
    def _list_statistics(pipe_output):
        data, *statistics = tuple(zip(*pipe_output))
        statistics = tuple(
            dict(enumerate(statistic)) for statistic in statistics
        )
        return (list(data),) + statistics


def _collect(comp, method, data):
    out = method(data)
    statistics = comp.get_latest_statistics()
    return out, statistics


def _collector_fun(comp, method):
    return lambda data: _collect(comp, method, data)


def is_empty_data(data):
    if isinstance(data, pc.data.Data) and data.n_rows == 0:
        return True
    if isinstance(data, np.ndarray) and len(data) == 0:
        return True
    if isinstance(data, pc.WeightedPoints):
        return is_empty_data(data.data)
    return False


def _weigh_and_filter(data, weigher, filter_, weighing, filtering):
    if is_empty_data(data):
        return (pc.WeightedPoints(data, []), {}, {})
    if weighing:
        weights, weigher_statistics = _collect(weigher, weigher.weigh, data)
    else:
        def_weigher = pc.AllOneWeigher()
        weights, weigher_statistics = _collect(
            def_weigher, def_weigher.weigh, data
        )

    weighed_data = pc.WeightedPoints(data, weights)

    filtered_data, filter_statistics = (
        _filter_data(filter_, weighed_data)
        if filtering
        else (weighed_data, {})
    )

    return filtered_data, weigher_statistics, filter_statistics


def _filter_data(filter_, weighted_points):
    points_to_filter, filter_statistics = _collect(
        filter_, filter_.filter, weighted_points.weights
    )
    return weighted_points[points_to_filter], filter_statistics


def _add_zeros(weighted_points, n_zeros):
    ws = weighted_points.points[:, 0]
    ws = np.linspace(min(ws), max(ws), n_zeros)

    zero = np.zeros(n_zeros)
    full = 360 * np.ones(n_zeros)
    zeros = np.column_stack((ws, zero, zero))
    fulls = np.column_stack((ws, full, zero))

    original_points = weighted_points.points
    original_weights = weighted_points.weights

    return pc.WeightedPoints(
        data=np.concatenate([original_points, zeros, fulls]),
        weights=np.concatenate([original_weights, np.ones(2 * n_zeros)]),
    )



# pylint: disable=missing-module-docstring

import csv
from ast import literal_eval
from inspect import getmembers, isfunction

import numpy as np

import hrosailing.pipelinecomponents.modelfunctions as model

from ._basepolardiagram import (
    PolarDiagram,
    PolarDiagramException,
    PolarDiagramInitializationException,
)
from ._plotting import (
    plot_color_gradient,
    plot_convex_hull,
    plot_flat,
    plot_polar,
    plot_surface,
)

MODEL_FUNCTIONS = dict(getmembers(model, isfunction))


class PolarDiagramCurve(PolarDiagram):
    """A class to represent, visualize and work with a polar diagram
    given by a fitted curve/surface.

    Parameters
    ----------
    f : function
        Curve/surface that describes the polar diagram, given as
        a function, with the signature `f(ws, wa, *params) -> bsp`,
        where

        - `ws` and `wa` should be `array_like` of shape `(n,)`
        and refer to wind speeds and wind angles,
        - `params` is a list of additional parameters,
        - `bsp` is `array_like` of shape `(n,)` and contains the resulting
        boat speeds.

    params : Sequence
        Optimal parameters for `f`.

    radians : bool, optional
        Specifies if `f` takes the wind angles in radians or degrees.

        Defaults to `False`.

    Raises
    ------
    PolarDiagramInitializationException
        If `f` is not callable.
    PolarDiagramInitializationException
        If `params` contains not enough parameters for `f`.
    """

    def __init__(self, f, *params, radians=False):
        if not callable(f):
            raise PolarDiagramInitializationException("`f` is not callable")

        if not self._check_enough_params(f, params):
            raise PolarDiagramInitializationException(
                "`params` is an incorrect amount of parameters for `f`"
            )

        self._f = f
        self._params = params
        self._rad = radians

    @staticmethod
    def _check_enough_params(func, params):
        try:
            func(1, 1, *params)
            return True
        except (IndexError, TypeError):
            return False

    def __repr__(self):
        return (
            f"PolarDiagramCurve(f={self._f.__name__},"
            f"{self._params}, radians={self._rad})"
        )

    def __call__(self, ws, wa):
        # do not change the input wa
        if isinstance(wa, np.ndarray):
            wa = wa.copy()

        if np.any((ws < 0)):
            raise PolarDiagramException("`ws` is negative")

        if self.radians:
            wa = np.rad2deg(wa)
            wa %= 360
            wa = np.deg2rad(wa)
        else:
            wa %= 360

        return self.curve(ws, wa, *self.parameters)

    @property
    def curve(self):
        """Returns a read only version of `self._f`."""
        return self._f

    @property
    def parameters(self):
        """Returns a read only version of `self._params`."""
        return self._params

    @property
    def radians(self):
        """Returns a read only version of `self._rad`."""
        return self._rad

    def to_csv(self, csv_path):
        """Creates a .csv file with delimiter ':' and the
        following format:

            `PolarDiagramCurve`
            Function: `self.curve.__name__`
            Radians: `self.radians`
            Parameters: `self.parameters`

        Parameters
        ----------
        csv_path : path-like
            Path to a .csv file or where a new .csv file will be created.

        Raises
        ------
        OSError
            If no permission to write is given for file.
        """
        with open(csv_path, "w", newline="", encoding="utf-8") as file:
            csv_writer = csv.writer(file, delimiter=":")
            csv_writer.writerow([self.__class__.__name__])
            csv_writer.writerow(["Function"] + [self.curve.__name__])
            csv_writer.writerow(["Radians"] + [str(self.radians)])
            csv_writer.writerow(["Parameters"] + list(self.parameters))

    @classmethod
    def __from_csv__(cls, file):
        csv_reader = csv.reader(file, delimiter=":")
        function = next(csv_reader)[1]
        radians = literal_eval(next(csv_reader)[1])
        params = [literal_eval(value) for value in next(csv_reader)[1:]]

        if function not in MODEL_FUNCTIONS:
            raise PolarDiagramInitializationException(
                f"no valid function, named {function}"
            )

        function = MODEL_FUNCTIONS[function]
        return PolarDiagramCurve(function, *params, radians=radians)

    def symmetrize(self):
        """Constructs a symmetric version of the polar diagram,
        by mirroring it at the 0° - 180° axis and returning a new instance.
        """

        def sym_func(ws, wa, *params):
            y = self._f(ws, wa, *params)
            y_symm = self._f(ws, (360 - wa) % 360, *params)
            return 0.5 * (y + y_symm)

        return PolarDiagramCurve(
            sym_func, *self.parameters, radians=self.radians
        )

    # TODO Add positivity checks for ws in various cases
    def get_slices(self, ws=None, n_steps=None):
        """For given wind speeds, return the slices of the polar diagram
        corresponding to them.

        Slices are equal to `self(w, wa)` where `w` goes through
        the given values in `ws` and `wa` goes through a fixed
        number of angles between 0° and 360°.

        Parameters
        ----------
        ws : tuple of length 2, iterable, int or float, optional
            Slices of the polar diagram given as either:

            - a tuple of length 2, specifying an interval of considered
            wind speeds. The amount of slices taken from that interval
            are determined by the parameter `n_steps`,
            - an iterable of specific wind speeds,
            - a single wind speed.

            If nothing is passed, it will default to `(0, 20)`.

        n_steps : positive int or float, optional
            Specifies the amount of slices taken from the given
            wind speed interval.

            Will only be used if `ws` is a tuple of length 2.

            If nothing is passed, it will default to `ws[1] - ws[0]`.

        Returns
        -------
        slices : tuple
            Slices of the polar diagram, given as a tuple of length 3,
            consisting of the given wind speeds `ws`, `self.wind_angles`
            (in rad) and a list of arrays containing the
            corresponding boat speeds.

        Raises
        ------
        PolarDiagramException
            If `n_steps` is non-positive.
        """
        if ws is None:
            ws = (0, 20)

        if isinstance(ws, (int, float)):
            ws = [ws]
        elif isinstance(ws, tuple) and len(ws) == 2:
            if n_steps is None:
                n_steps = int(round(ws[1] - ws[0]))

            if n_steps <= 0:
                raise PolarDiagramException("`n_steps` is non-positive")

            ws = list(np.linspace(ws[0], ws[1], n_steps))

        wa = np.linspace(0, 360, 1000)
        if self.radians:
            wa = np.deg2rad(wa)

        bsp = [self(np.array([w] * 1000), wa) for w in ws]

        if not self.radians:
            wa = np.deg2rad(wa)

        return ws, wa, bsp

    # pylint: disable=arguments-renamed
    def plot_polar(
        self,
        ws=None,
        n_steps=None,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """Creates a polar plot of one or more slices of the polar diagram.

        Parameters
        ----------
        ws : tuple of length 2, iterable, int or float, optional
            Slices of the polar diagram given as either:

            - a tuple of length 2, specifying an interval of considered
            wind speeds. The amount of slices taken from that interval
            are determined by the parameter `n_steps`,
            - an iterable of specific wind speeds,
            - a single wind speed.

            Slices will then equal `self(w, wa)` where `w` goes through
            the given values in `ws` and `wa` goes through a fixed
            number of angles between 0° and 360°.

            If nothing is passed, it will default to `(0, 20)`.

        n_steps : positive int or float, optional
            Specifies the amount of slices taken from the given
            wind speed interval.

            Will only be used if `ws` is a tuple of length 2.

            If nothing is passed, it will default to `ws[1] - ws[0]`.

        ax : matplotlib.projections.polar.PolarAxes, optional
            Axes instance where the plot will be created.

        colors : color_like or sequence of color_likes or (ws, color_like) pairs, optional
            Specifies the colors to be used for the different slices.

            - If a color_like is passed, all slices will be plotted in the
            respective color.
            - If 2 colors are passed, slices will be plotted with a color
            gradient that is determined by the corresponding wind speed.
            - Otherwise the slices will be colored in turn with the specified
            colors or the color `"blue"`, if there are too few colors. The
            order is determined by the corresponding wind speeds.
            - Alternatively one can specify certain slices to be plotted in
            a color out of order by passing a sequence of `(ws, color)` pairs.

            Defaults to `("green", "red")`.

        show_legend : bool, optional
            Specifies whether a legend will be shown next to the plot.

            The type of legend depends on the color options.

            If plotted with a color gradient, a `matplotlib.colorbar.Colorbar`
            will be created, otherwise a `matplotlib.legend.Legend` instance.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.

            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.

            Will only be used if `show_legend` is `True`.

        plot_kw : Keyword arguments
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.
        """
        ws, wa, bsp = self.get_slices(ws, n_steps)
        wa = [wa] * len(ws)

        plot_polar(
            ws,
            wa,
            bsp,
            ax,
            colors,
            show_legend,
            legend_kw,
            _lines=True,
            **plot_kw,
        )

    # pylint: disable=arguments-renamed
    def plot_flat(
        self,
        ws=None,
        n_steps=None,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """Creates a cartesian plot of one or more slices of the polar diagram.

        Parameters
        ----------
        ws : tuple of length 2, iterable, int or float, optional
            Slices of the polar diagram given as either:

            - a tuple of length 2, specifying an interval of considered
            wind speeds. The amount of slices taken from that interval
            are determined by the parameter `n_steps`,
            - an iterable of specific wind speeds,
            - a single wind speed.

            Slices will then equal `self(w, wa)` where `w` goes through
            the given values in `ws` and `wa` goes through a fixed
            number of angles between 0° and 360°.

            If nothing is passed, it will default to `(0, 20)`.

        n_steps : positive int or float, optional
            Specifies the amount of slices taken from the given
            wind speed interval.

            Will only be used if `ws` is a tuple of length 2.

            If nothing is passed, it will default to `ws[1] - ws[0]`.

        ax : matplotlib.axes.Axes, optional
            Axes instance where the plot will be created.

        colors : color_like or sequence of color_likes or (ws, color_like) pairs, optional
            Specifies the colors to be used for the different slices.

            - If a color_like is passed, all slices will be plotted in the
            respective color.
            - If 2 colors are passed, slices will be plotted with a color
            gradient that is determined by the corresponding wind speed.
            - Otherwise the slices will be colored in turn with the specified
            colors or the color `"blue"`, if there are too few colors. The
            order is determined by the corresponding wind speeds.
            - Alternatively one can specify certain slices to be plotted in
            a color out of order by passing a sequence of `(ws, color)` pairs.

            Defaults to `("green", "red")`.

        show_legend : bool, optional
            Specifies whether a legend will be shown next to the plot.

            The type of legend depends on the color options.

            If plotted with a color gradient, a `matplotlib.colorbar.Colorbar`
            will be created, otherwise a `matplotlib.legend.Legend` instance.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.

            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.

            Will only be used if show_legend is `True`.

        plot_kw : Keyword arguments
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.
        """
        ws, wa, bsp = self.get_slices(ws, n_steps)
        wa = [np.rad2deg(wa)] * len(ws)

        plot_flat(
            ws,
            wa,
            bsp,
            ax,
            colors,
            show_legend,
            legend_kw,
            _lines=True,
            **plot_kw,
        )

    def plot_3d(self, ws=None, n_steps=None, ax=None, colors=("green", "red")):
        """Creates a 3d plot of a part of the polar diagram.

        Parameters
        ----------
        ws : tuple of length 2, optional
            A region of the polar diagram given as an interval of
            wind speeds.

            Slices will then equal `self(w, wa)` where `w` goes through
            the given values in `ws` and `wa` goes through a fixed
            number of angles between 0° and 360°.

            If nothing is passed, it will default to `(0, 20)`.

        n_steps : positive int or float, optional
            Specifies the amount of slices taken from the given
            interval in `ws`.

            If nothing is passed, it will default to `100`.

        ax : mpl_toolkits.mplot3d.axes3d.Axes3D, optional
            Axes instance where the plot will be created.

        colors: tuple of two (2) color_likes, optional
            Color pair determining the color gradient with which the
            polar diagram will be plotted.

            Will be determined by the corresponding wind speeds.

            Defaults to `("green", "red")`.
        """
        if n_steps is None:
            n_steps = 100

        ws, wa, bsp = self.get_slices(ws, n_steps)
        bsp = np.array(bsp).T
        ws, wa = np.meshgrid(ws, wa)
        bsp, wa = bsp * np.cos(wa), bsp * np.sin(wa)

        plot_surface(ws, wa, bsp, ax, colors)

    def plot_color_gradient(
        self,
        ws=None,
        n_steps=None,
        ax=None,
        colors=("green", "red"),
        marker=None,
        ms=None,
        show_legend=False,
        **legend_kw,
    ):
        """Creates a 'wind speed vs. wind angle' color gradient plot
        of a part of the polar diagram with respect to the corresponding
        boat speeds.

        Parameters
        ----------
        ws :  tuple of length 3, optional
            A region of the polar diagram given as an interval of
            wind speeds.

            Slices will then equal `self(w, wa)` where `w` goes through
            the given values in `ws` and `wa` goes through a fixed
            number of angles between 0° and 360°.

            If nothing is passed, it will default to `(0, 20)`.

        n_steps : positive int or float, optional
            Specifies the amount of slices taken from the given
            interval in `ws`.

            If nothing is passed, it will default to `100`.

        ax : matplotlib.axes.Axes, optional
            Axes instance where the plot will be created.

        colors : tuple of two (2) color_likes, optional
            Color pair determining the color gradient with which the
            polar diagram will be plotted.

            Will be determined by the corresponding boat speed.

            Defaults to `("green", "red")`.

        marker : matplotlib.markers.Markerstyle or equivalent, optional
            Markerstyle for the created scatter plot.

            Defaults to `"o"`.

        ms : float or array_like of fitting shape, optional
            Marker size in points**2.

        show_legend : bool, optional
            Specifies whether a legend will be shown next
            to the plot.

            Legend will be a `matplotlib.colorbar.Colorbar` instance.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.

            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.

            Will only be used if `show_legend` is `True`.
        """
        if n_steps is None:
            n_steps = 100

        if isinstance(ws, set):
            ws = sorted(list(ws))

        ws, wa, bsp = self.get_slices(ws, n_steps)
        wa = np.rad2deg(wa)
        ws, wa = np.meshgrid(ws, wa)
        bsp = np.array(bsp).T

        plot_color_gradient(
            ws.ravel(),
            wa.ravel(),
            bsp.ravel(),
            ax,
            colors,
            marker,
            ms,
            show_legend,
            **legend_kw,
        )

    # pylint: disable=arguments-renamed
    def plot_convex_hull(
        self,
        ws=None,
        n_steps=None,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """Computes the (separate) convex hull of one or more
        slices of the polar diagram and creates a polar plot of them.

        Parameters
        ----------
        ws : tuple of length 2, iterable, int or float, optional
            Slices of the polar diagram given as either:

            - a tuple of length 2, specifying an interval of considered
            wind speeds. The amount of slices taken from that interval
            are determined by the parameter `n_steps`,
            - an iterable of specific wind speeds,
            - a single wind speed.

            Slices will then equal `self(w, wa)` where `w` goes through
            the given values in `ws` and `wa` goes through a fixed
            number of angles between 0° and 360°.

            If nothing is passed, it will default to `(0, 20)`.

        n_steps : positive int or float, optional
            Specifies the amount of slices taken from the given
            wind speed interval.

            Will only be used if `ws` is a tuple of length 2.

            If nothing is passed, it will default to `ws[1] - ws[0]`.

        ax : matplotlib.projections.polar.PolarAxes, optional
            Axes instance where the plot will be created.

        colors : sequence of color_likes or (ws, color_like) pairs, optional
            Specifies the colors to be used for the different slices.

            - If 2 colors are passed, slices will be plotted with a color
            gradient that is determined by the corresponding wind speed.
            - Otherwise the slices will be colored in turn with the specified
            colors or the color `"blue"`, if there are too few colors. The
            order is determined by the corresponding wind speeds.
            - Alternatively one can specify certain slices to be plotted in
            a color out of order by passing a sequence of `(ws, color)` pairs.

            Defaults to `("green", "red")`.

        show_legend : bool, optional
            Specifies whether or not a legend will be shown next to the plot.

            The type of legend depends on the color options.

            If plotted with a color gradient, a `matplotlib.colorbar.Colorbar`
            will be created, otherwise a `matplotlib.legend.Legend`.

            Defaults to `False`.

        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.

            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.

            Will only be used if `show_legend` is `True`.

        plot_kw : Keyword arguments
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.
        """
        ws, wa, bsp = self.get_slices(ws, n_steps)
        wa = [wa] * len(ws)

        plot_convex_hull(
            ws,
            wa,
            bsp,
            ax,
            colors,
            show_legend,
            legend_kw,
            _lines=True,
            **plot_kw,
        )



# pylint: disable=missing-module-docstring
# pylint: disable=too-many-lines

import csv
import enum
import warnings
from ast import literal_eval
from typing import Iterable

import numpy as np

from hrosailing.pipelinecomponents import (
    ArithmeticMeanInterpolator,
    Ball,
    WeightedPoints,
)

from ._basepolardiagram import (
    PolarDiagram,
    PolarDiagramException,
    PolarDiagramInitializationException,
)
from ._plotting import (
    plot_color_gradient,
    plot_convex_hull,
    plot_flat,
    plot_polar,
    plot_surface,
)


class PolarDiagramTable(PolarDiagram):
    """A class to represent, visualize and work with a polar diagram
    in the form of a table.

    Parameters
    ----------
    ws_resolution : array_like, default: `numpy.arange(2, 42, 2)`
        Wind speeds that will correspond to the columns of the table.
        - If array_like, resolution will be `numpy.array(ws_resolution)`.
        - If a scalar `num`, resolution will be `numpy.arange(num, 40, num)`.
    wa_resolution : array_like, default: `numpy.arange(0, 360, 5)`
        Wind angles that will correspond to the rows of the table.
        - If array_like, resolution will be `numpy.array(wa_resolution)`.
        - If a scalar `num`, resolution will be `numpy.arange(num, 360, num)`.
    bsps : array_like, default: `numpy.zeros((rdim, cdim))`
        Boat speeds that will correspond to the entries of the table.
        Needs to have dimensions matching `ws_resolution` and `wa_resolution`.

    Attributes
    ----------
    wind_speeds
    wind_angles
    boat_speeds

    Examples
    --------
    >>> pd = PolarDiagramTable()
    >>> pd.wind_speeds
    [ 2  4  6  8 10 12 14 16 18 20 22 24 26 28 30 32 34 36 38 40]
    >>> pd.wind_angles
    [  0   5  10  15  20  25  30  35  40  45  50  55  60  65  70  75  80
      85  90  95 100 105 110 115 120 125 130 135 140 145 150 155 160 165
     170 175 180 185 190 195 200 205 210 215 220 225 230 235 240 245 250
     255 260 265 270 275 280 285 290 295 300 305 310 315 320 325 330 335
     340 345 350 355]
    >>> pd = PolarDiagramTable(
    ...     bsps=[
    ...         [5.33, 6.32, 6.96, 7.24, 7.35],
    ...         [5.64, 6.61, 7.14, 7.42, 7.56],
    ...         [5.89, 6.82, 7.28, 7.59, 7.84],
    ...         [5.92, 6.98, 7.42, 7.62, 7.93],
    ...         [5.98, 7.07, 7.59, 8.02, 8.34],
    ...         [5.8, 6.95, 7.51, 7.98, 8.52],
    ...         [5.2, 6.41, 7.19, 7.66, 8.14]
    ...     ],
    ...     ws_resolution=[6, 8, 10, 12, 14],
    ...     wa_resolution=[52, 60, 75, 90, 110, 120, 135],
    ... )
    >>> print(pd)
      TWA / TWS    6.0    8.0    10.0    12.0    14.0
    +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
    52.0          5.33   6.32    6.96    7.24    7.35
    60.0          5.64   6.61    7.14    7.42    7.56
    75.0          5.89   6.82    7.28    7.59    7.84
    90.0          5.92   6.98    7.42    7.62    7.93
    110.0         5.98   7.07    7.59    8.02    8.34
    120.0         5.80   6.95    7.51    7.98    8.52
    135.0         5.20   6.41    7.19    7.66    8.14

    Single entries can be extracted like so
    >>> pd[10, 90]
    7.42

    or if a wind speed (resp. wind angle) isn't in the table
    a value can be interpolated
    >>> pd(11, 90)
    7.50924383603392
    """

    def __init__(self, ws_resolution=None, wa_resolution=None, bsps=None):
        ws_resolution = _Resolution.WIND_SPEED.set_resolution(ws_resolution)
        wa_resolution = _Resolution.WIND_ANGLE.set_resolution(wa_resolution)

        if bsps is None:
            self._create_zero_table(ws_resolution, wa_resolution)
            return

        bsps = np.asarray_chkfinite(bsps, float)

        # sanity checks
        if bsps.dtype is object:
            raise PolarDiagramInitializationException(
                "`bsps` is not array_like"
            )
        if _incompatible_shapes(bsps, ws_resolution, wa_resolution):
            raise PolarDiagramInitializationException(
                "`bsps` has incorrect shape"
            )

        (
            self._ws_resolution,
            self._wa_resolution,
            self._boat_speeds,
        ) = _sort_table(ws_resolution, wa_resolution, bsps)

    def _create_zero_table(self, ws_resolution, wa_resolution):
        rows, cols = len(wa_resolution), len(ws_resolution)
        self._boat_speeds = np.zeros((rows, cols))
        self._ws_resolution = sorted(ws_resolution)
        self._wa_resolution = sorted(wa_resolution)

    def __str__(self):
        table = ["  TWA / TWS"]
        ws = self.wind_speeds
        if len(ws) <= 15:
            self._create_short_table(table, ws)
        else:
            wind = []
            wind.extend(ws[:5])
            wind.extend(ws[-5:])
            self._create_long_table(table, wind)

        return "".join(table)

    def _create_short_table(self, table, wind):
        bsps = self.boat_speeds

        table.extend([f"    {float(ws):.1f}" for ws in wind])
        table.append("\n+++++++++++")

        for ws in wind:
            le = len(f"{float(ws):.1f}")
            table.append("  ".ljust(le + 4, "+"))

        table.append("\n")

        for i, wa in enumerate(self.wind_angles):
            table.append(f"{float(wa):.1f}".ljust(11))

            for j, ws in enumerate(wind):
                le = len(f"{float(ws):.1f}")
                table.append(f"{bsps[i][j]:.2f}".rjust(4 + le))

            table.append("\n")

    def _create_long_table(self, table, wind):
        bsps = self.boat_speeds

        for i, ws in enumerate(wind):
            if i == 5:
                table.append("  ...")

            table.append(f"    {float(ws):.1f}")

        table.append("\n+++++++++++")

        for i, ws in enumerate(wind):
            if i == 5:
                table.append("  +++")

            le = len(f"{float(ws):.1f}")
            table.append("  ".ljust(le + 4, "+"))

        table.append("\n")

        for i, wa in enumerate(self.wind_angles):
            table.append(f"{float(wa):.1f}".rjust(11))

            for j, ws in enumerate(wind):
                if j == 5:
                    table.append("  ...")

                le = len(f"{float(ws):.1f}")
                table.append(f"{bsps[i][j]:.2f}".rjust(4 + le))

            table.append("\n")

    def __repr__(self):
        return (
            f"PolarDiagramTable({self.wind_speeds}, {self.wind_angles}, "
            f"{self.boat_speeds})"
        )

    def __call__(
        self,
        ws,
        wa,
        interpolator=ArithmeticMeanInterpolator(params=(50,)),
        neighbourhood=Ball(radius=1),
    ):
        """Calculates the boat speed for given `ws` and `wa`.

        If the `ws-wa` point is in the table, the corresponding entry is
        returned, otherwise the value is interpolated.

        Parameters
        ----------
        ws : int or float
            Wind speed.

        wa : int or float
            Wind angle.

        interpolator : hrosailing.pipelinecomponents.Interpolator, optional
            Interpolator subclass that determines the interpolation
            method used to determine the value at the ws-wa point.

            Defaults to: `hrosailing.pipelinecomponents.ArithmeticMeanInterpolator(params = (50, ))`

        neighbourhood : hrosailing.pipelinecomponents.Neighbourhood, default: `hrosailing.pipelinecomponents.Ball(1)`
            Neighbourhood subclass used to determine the grid points
            in the table that will be used in the interpolation.

        Returns
        -------
        bsp : int or float
            Boat speed value as determined above.
        """
        try:
            return self[ws, wa]
        except PolarDiagramException:
            point = np.array([ws, wa])
            ws, wa = np.meshgrid(self.wind_speeds, self.wind_angles)
            points = np.column_stack(
                (ws.ravel(), wa.ravel(), self.boat_speeds.ravel())
            )

            weighted_points = WeightedPoints(points, weights=1)

            considered_points = neighbourhood.is_contained_in(
                points[:, :2] - point
            )

            return interpolator.interpolate(
                weighted_points[considered_points], point
            )

    def __getitem__(self, *key):
        """Returns the value of a given entry in the table."""
        ws, wa = key[0]

        col = self._get_indices(np.atleast_1d(ws), _Resolution.WIND_SPEED)
        row = self._get_indices(np.atleast_1d(wa), _Resolution.WIND_ANGLE)
        return float(self.boat_speeds[row, col])

    def _get_indices(self, wind, soa):
        res = (
            self.wind_speeds
            if soa.name == _Resolution.WIND_SPEED.name
            else self.wind_angles
        )

        if wind is None:
            return range(len(res))

        wind = soa.normalize_wind(wind)

        # sanity checks
        if not wind:
            raise PolarDiagramException("empty slice-list was passed")

        if not wind.issubset(set(res)):
            raise PolarDiagramException(f"{wind} is not contained in {res}")

        return [i for i, w in enumerate(res) if w in wind]

    @property
    def wind_angles(self):
        """Returns a read only version of `self._wa_resolution`."""
        return self._wa_resolution.copy()

    @property
    def wind_speeds(self):
        """Returns a read only version of `self._ws_resolution`."""
        return self._ws_resolution.copy()

    @property
    def boat_speeds(self):
        """Returns a read only version of `self._boat_speeds`."""
        return self._boat_speeds.copy()

    def to_csv(self, csv_path, fmt="hro"):
        """Creates a .csv file with delimiter ',' and the following format:

            `PolarDiagramTable`
            TWS:
            `self.wind_speeds`
            TWA:
            `self.wind_angles`
            Boat speeds:
            `self.boat_speeds`

        Parameters
        ----------
        csv_path : path_like
            Path for the .csv file.
        fmt : {"hro", "orc", "opencpn", "array"}, default: "hro"
            Format for the .csv file.

        See Also
        --------
        `from_csv`

        Examples
        --------
        >>> pd = PolarDiagramTable(
        ...     bsps=[
        ...         [5.33, 6.32, 6.96, 7.24, 7.35],
        ...         [5.64, 6.61, 7.14, 7.42, 7.56],
        ...         [5.89, 6.82, 7.28, 7.59, 7.84],
        ...         [5.92, 6.98, 7.42, 7.62, 7.93],
        ...         [5.98, 7.07, 7.59, 8.02, 8.34],
        ...         [5.8, 6.95, 7.51, 7.98, 8.52],
        ...         [5.2, 6.41, 7.19, 7.66, 8.14]
        ...     ],
        ...     ws_resolution=[6, 8, 10, 12, 14],
        ...     wa_resolution=[52, 60, 75, 90, 110, 120, 135],
        ... )
        >>> print(pd)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0
        +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
        52.0          5.33   6.32    6.96    7.24    7.35
        60.0          5.64   6.61    7.14    7.42    7.56
        75.0          5.89   6.82    7.28    7.59    7.84
        90.0          5.92   6.98    7.42    7.62    7.93
        110.0         5.98   7.07    7.59    8.02    8.34
        120.0         5.80   6.95    7.51    7.98    8.52
        135.0         5.20   6.41    7.19    7.66    8.14
        >>> pd.to_csv("example.csv")
        >>> pd2 = from_csv("example.csv")
        >>> print(pd2)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0
        +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
        52.0          5.33   6.32    6.96    7.24    7.35
        60.0          5.64   6.61    7.14    7.42    7.56
        75.0          5.89   6.82    7.28    7.59    7.84
        90.0          5.92   6.98    7.42    7.62    7.93
        110.0         5.98   7.07    7.59    8.02    8.34
        120.0         5.80   6.95    7.51    7.98    8.52
        135.0         5.20   6.41    7.19    7.66    8.14
        """
        if fmt not in {"hro", "orc", "opencpn", "array"}:
            raise PolarDiagramException("`fmt` not implemented")

        with open(csv_path, "w", newline="", encoding="utf-8") as file:
            if fmt == "orc":
                self._write_orc_format(file)
                return

            if fmt == "array":
                self._write_array_format(file)
                return

            csv_writer = csv.writer(file, delimiter=",")

            if fmt == "opencpn":
                self._write_opencpn_format(csv_writer)
                return

            self._write_hro_format(csv_writer)

    def _write_orc_format(self, file):
        csv_writer = csv.writer(file, delimiter=";")
        csv_writer.writerow(["twa/tws"] + list(self.wind_speeds))
        csv_writer.writerow([0] * (len(self.wind_speeds) + 1))

        self._write_rows(csv_writer)

    def _write_rows(self, csv_writer):
        rows = np.column_stack((self.wind_angles, self.boat_speeds))
        csv_writer.writerows(rows)

    def _write_array_format(self, file):
        csv_writer = csv.writer(file, delimiter="\t")
        csv_writer.writerow([r"TWA\TWS"] + list(self.wind_speeds))

        self._write_rows(csv_writer)

    def _write_opencpn_format(self, csv_writer):
        csv_writer.writerow([r"TWA \ TWS"] + list(self.wind_speeds))

        self._write_rows(csv_writer)

    def _write_hro_format(self, csv_writer):
        csv_writer.writerow([self.__class__.__name__])
        csv_writer.writerow(["TWS"])
        csv_writer.writerow(self.wind_speeds)
        csv_writer.writerow(["TWA"])
        csv_writer.writerow(self.wind_angles)
        csv_writer.writerow(["BSP"])
        csv_writer.writerows(self.boat_speeds)

    @classmethod
    def __from_csv__(cls, file):
        csv_reader = csv.reader(file, delimiter=",")
        next(csv_reader)
        ws_res = [literal_eval(ws) for ws in next(csv_reader)]
        next(csv_reader)
        wa_res = [literal_eval(wa) for wa in next(csv_reader)]
        next(csv_reader)
        bsps = [[literal_eval(bsp) for bsp in row] for row in csv_reader]

        return PolarDiagramTable(ws_res, wa_res, bsps)

    def symmetrize(self):
        """Constructs a symmetric version of the polar diagram.

        Symmetrizes polar diagram by mirroring it at the 0° - 180° axis
        and returning a new instance.

        Warnings
        --------
        Should only be used if all the wind angles of the initial
        polar diagram are on one side of the 0° - 180° axis,
        otherwise this can lead to duplicate data, which can
        overwrite or live alongside old data.

        Examples
        --------
        >>> pd = PolarDiagramTable(
        ...     bsps=[
        ...         [5.33, 6.32, 6.96, 7.24, 7.35],
        ...         [5.64, 6.61, 7.14, 7.42, 7.56],
        ...         [5.89, 6.82, 7.28, 7.59, 7.84],
        ...         [5.92, 6.98, 7.42, 7.62, 7.93],
        ...     ],
        ...     ws_resolution = [6, 8, 10, 12, 14],
        ...     wa_resolution = [52, 60, 75, 90]
        ... )
        >>> print(pd)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0
        +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
        52.0          5.33   6.32    6.96    7.24    7.35
        60.0          5.64   6.61    7.14    7.42    7.56
        75.0          5.89   6.82    7.28    7.59    7.84
        90.0          5.92   6.98    7.42    7.62    7.93
        >>> sym_pd = pd.symmetrize()
        >>> print(sym_pd)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0
        +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
        52.0          5.33   6.32    6.96    7.24    7.35
        60.0          5.64   6.61    7.14    7.42    7.56
        75.0          5.89   6.82    7.28    7.59    7.84
        90.0          5.92   6.98    7.42    7.62    7.93
        270.0         5.92   6.98    7.42    7.62    7.93
        285.0         5.89   6.82    7.28    7.59    7.84
        300.0         5.64   6.61    7.14    7.42    7.56
        308.0         5.33   6.32    6.96    7.24    7.35
        """

        symmetric_wa_resolution = np.concatenate(
            [self.wind_angles, 360 - np.flip(self.wind_angles)]
        )
        symmetric_bsps = np.row_stack(
            (self.boat_speeds, np.flip(self.boat_speeds, axis=0))
        )

        if 180 in self.wind_angles:
            (
                symmetric_wa_resolution,
                symmetric_bsps,
            ) = _delete_multiple_180_degree_occurrences(
                symmetric_wa_resolution, symmetric_bsps
            )
        if 0 in self.wind_angles:
            (
                symmetric_wa_resolution,
                symmetric_bsps,
            ) = _delete_multiple_0_degree_occurrences(
                symmetric_wa_resolution, symmetric_bsps
            )

        return PolarDiagramTable(
            ws_resolution=self.wind_speeds,
            wa_resolution=symmetric_wa_resolution,
            bsps=symmetric_bsps,
        )

    def change_entries(self, new_bsps, ws=None, wa=None):
        """Changes specified entries in the table.

        Parameters
        ----------
        new_bsps: array_like of matching shape
            Sequence containing the new boat speeds to be inserted
            in the specified entries.
        ws: array_like, default: `self.wind_speeds`
            Columns of table given as elements of `self.wind_speed`.
        wa: array_like, default: `self.wind_angles`
            Rows of table given as elements of `self.wind_angles`.

        Examples
        --------
        >>> pd = PolarDiagramTable(
        ...     ws_resolution=[6, 8, 10, 12, 14],
        ...     wa_resolution=[52, 60, 75, 90, 110, 120, 135]
        ... )
        >>> print(pd)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0
        +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
        52.0          0.00   0.00    0.00    0.00    0.00
        60.0          0.00   0.00    0.00    0.00    0.00
        75.0          0.00   0.00    0.00    0.00    0.00
        90.0          0.00   0.00    0.00    0.00    0.00
        110.0         0.00   0.00    0.00    0.00    0.00
        120.0         0.00   0.00    0.00    0.00    0.00
        135.0         0.00   0.00    0.00    0.00    0.00
        >>> pd.change_entries(
        ...     new_bsps=[5.33, 5.64, 5.89, 5.92, 5.98, 5.8, 5.2],
        ...     ws=6
        ... )
        >>> print(pd)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0
        +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
        52.0          5.33   0.00    0.00    0.00    0.00
        60.0          5.64   0.00    0.00    0.00    0.00
        75.0          5.89   0.00    0.00    0.00    0.00
        90.0          5.92   0.00    0.00    0.00    0.00
        110.0         5.98   0.00    0.00    0.00    0.00
        120.0         5.80   0.00    0.00    0.00    0.00
        135.0         5.20   0.00    0.00    0.00    0.00
        >>> pd.change_entries(
        ...     new_bsps=[5.7, 6.32, 6.96, 7.24, 7.35],
        ...     wa=52
        ... )
        >>> print(pd)
          TWA / TWS    6.0    8.0    10.0    12.0    14.0
        +++++++++++  +++++  +++++  ++++++  ++++++  ++++++
        52.0          5.70   6.32    6.96    7.24    7.35
        60.0          5.64   0.00    0.00    0.00    0.00
        75.0          5.89   0.00    0.00    0.00    0.00
        90.0          5.92   0.00    0.00    0.00    0.00
        110.0         5.98   0.00    0.00    0.00    0.00
        120.0         5.80   0.00    0.00    0.00    0.00
        135.0         5.20   0.00    0.00    0.00    0.00
        """
        new_bsps = np.atleast_1d(new_bsps)

        new_bsps = np.asarray_chkfinite(new_bsps)

        if new_bsps.dtype == object:
            raise PolarDiagramException("`new_bsps` is not array_like")

        ws = self._get_indices(ws, _Resolution.WIND_SPEED)
        wa = self._get_indices(wa, _Resolution.WIND_ANGLE)

        wa_len = len(wa) == 1
        ws_len = len(ws) == 1

        # wrong shape can lead to missing assignments
        if wa_len and ws_len:
            correct_shape = new_bsps.shape == (1,)
        elif wa_len:
            correct_shape = new_bsps.shape == (len(ws),)
        elif ws_len:
            correct_shape = new_bsps.shape == (len(wa),)
        else:
            correct_shape = new_bsps.shape == (len(wa), len(ws))

        if not correct_shape:
            raise PolarDiagramException("`new_bsps` has wrong shape")

        mask = np.zeros(self.boat_speeds.shape, dtype=bool)
        for i in wa:
            for j in ws:
                mask[i, j] = True

        self._boat_speeds[mask] = new_bsps.flat

    def get_slices(self, ws=None):
        """Returns given slice of polar diagram.

        The slices are equal to the corresponding columns of the table
        together with `self.wind_angles`.

        Parameters
        ----------
        ws : array_like, default: `self.wind_speeds`
            Slices of the polar diagram table, given as either:
            - a tuple of length 2 specifying an interval of
                considered wind speeds,
            - an iterable containing only elements of `self.wind_speeds`,
            - a single element of `self.wind_speeds`.

        Returns
        -------
        slices : tuple
            Slices of the polar diagram, given as a tuple of length 3,
            consisting of the given wind speeds `ws`, `self.wind_angles`
            (in rad) and an array with the corresponding columns of the table.

        Raises
        ------
        PolarDiagramException
            If at least one element of `ws` is not in `self.wind_speeds`.
        PolarDiagramException
            If the given interval doesn't contain any slices of the
            polar diagram.
        """
        if ws is None:
            ws = self.wind_speeds
        elif isinstance(ws, (int, float)):
            ws = [ws]
        elif isinstance(ws, tuple) and len(ws) == 2:
            ws = [w for w in self.wind_speeds if ws[0] <= w <= ws[1]]

        ws = sorted(list(ws))
        if not ws:
            raise PolarDiagramException("no slices were given")

        ind = self._get_indices(ws, _Resolution.WIND_SPEED)
        wa = np.deg2rad(self.wind_angles)

        bsp = self.boat_speeds[:, ind]
        if max(wa) > np.deg2rad(350) and min(wa) < np.deg2rad(10):
            bsp = np.concatenate([bsp, [bsp[0]]])
            wa = np.concatenate([wa, [wa[0]]])
        return ws, wa, bsp

    def plot_polar(
        self,
        ws=None,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """Creates a polar plot of one or more slices of the polar diagram.

        Parameters
        ----------
        ws : array_like, default: `self.wind_speeds`
            Slices of the polar diagram table, given as either:
            - a tuple of length 2 specifying an interval of
                considered wind speeds,
            - an iterable containing only elements of `self.wind_speeds`,
            - a single element of `self.wind_speeds`.
            The slices are then equal to the corresponding
            columns of the table together with `self.wind_angles`.

        ax : matplotlib.projections.polar.PolarAxes, optional
            Axes instance where the plot will be created.

        colors : color_like or sequence of color_likes or (ws, color_like) pairs
            Specifies the colors to be used for the different slices.

            - If a color_like is passed, all slices will be plotted in the
            respective color.
            - If 2 colors are passed, slices will be plotted with a color
            gradient that is determined by the corresponding wind speed.
            - Otherwise the slices will be colored in turn with the specified
            colors or the color `"blue"`, if there are too few colors. The
            order is determined by the corresponding wind speeds.
            - Alternatively one can specify certain slices to be plotted in
            a color out of order by passing a sequence of `(ws, color)` pairs.

            Defaults to `("green", "red")`.

        show_legend : bool, default: `False`
            Specifies wether or not a legend will be shown next to the plot.

            The type of legend depends on the color options.

            If plotted with a color gradient, a `matplotlib.colorbar.Colorbar`
            will be created, otherwise a `matplotlib.legend.Legend` instance.

        Other Parameters
        ----------------
        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.
            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.
        plot_kw : dict, optional
            Keyword arguments to change various appearances of the plot.
            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.

        Raises
        ------
        PolarDiagramException
            If at least one element of `ws` is not in `self.wind_speeds`.
        PolarDiagramException
            If the given interval doesn't contain any slices of the
            polar diagram.

        Examples
        --------
        >>> import matplotlib.pyplot as pyplot
        >>> pd = from_csv("src/polar_diagrams/orc/A-35.csv", fmt="orc")
        >>> pd.plot_polar(
        ...     ws=[6, 8], show_legend=True, ls="-", lw=1.5, marker=""
        ... )
        >>> pyplot.show()

        ![logo](https://raw.githubusercontent.com/hrosailing/hrosailing/main/examples/pictures/table_plot_polar.png)
        """
        ws, wa, bsp = self.get_slices(ws)
        bsp = list(bsp.T)
        wa = [wa] * len(bsp)
        plot_polar(
            ws,
            wa,
            bsp,
            ax,
            colors,
            show_legend,
            legend_kw,
            _lines=True,
            **plot_kw,
        )

    def plot_flat(
        self,
        ws=None,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """Creates a cartesian plot of one or more slices of the polar diagram.

        Parameters
        ----------
        ws : array_like, default: `self.wind_speeds`
            Slices of the polar diagram table, given as either:
            - a tuple of length 2 specifying an interval of considered
            wind speeds,
            - an iterable containing only elements of `self.wind_speeds`,
            - a single element of `self.wind_speeds`.

            The slices are then equal to the corresponding
            columns of the table together with `self.wind_angles`.
        ax : matplotlib.axes.Axes, optional
            Axes instance where the plot will be created.
        colors : color_like or sequence of color_likes or (ws, color_like) pairs
            Specifies the colors to be used for the different slices.
            - If a color_like is passed, all slices will be plotted in the
            respective color.
            - If 2 colors are passed, slices will be plotted with a color
            gradient that is determined by the corresponding wind speed.
            - Otherwise the slices will be colored in turn with the specified
            colors or the color `"blue"`, if there are too few colors. The
            order is determined by the corresponding wind speeds.
            - Alternatively one can specify certain slices to be plotted in
            a color out of order by passing a sequence of `(ws, color)` pairs.

            Defaults to `("green", "red")`.

        show_legend : bool, default: `False`
            Specifies wether or not a legend will be shown next to the plot.

            The type of legend depends on the color options.

            If plotted with a color gradient, a `matplotlib.colorbar.Colorbar`
            will be created, otherwise a `matplotlib.legend.Legend` instance.
        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.
            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.
        plot_kw : dict, optional
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.

        Raises
        ------
        PolarDiagramException
            If at least one element of `ws` is not in `self.wind_speeds`.
        PolarDiagramException
            If the given interval doesn't contain any slices of the
            polar diagram.

        Examples
        --------
        >>> import matplotlib.pyplot as pyplot
        >>> pd = from_csv("src/polar_diagrams/orc/A-35.csv", fmt="orc")
        >>> pd.plot_flat(
        ...     ws=[6, 8], show_legend=True, ls="-", lw=1.5, marker=""
        ... )
        >>> pyplot.show()

        ![logo](https://raw.githubusercontent.com/hrosailing/hrosailing/main/examples/pictures/table_plot_flat.png)
        """
        ws, wa, bsp = self.get_slices(ws)
        bsp = list(bsp.T)
        wa = [np.rad2deg(wa)] * len(bsp)
        plot_flat(
            ws,
            wa,
            bsp,
            ax,
            colors,
            show_legend,
            legend_kw,
            _lines=True,
            **plot_kw,
        )

    def plot_3d(self, ax=None, colors=("green", "red")):
        """Creates a 3d plot of the polar diagram.

        Parameters
        ----------
        ax : mpl_toolkits.mplot3d.axes3d.Axes3D, optional
            Axes instance where the plot will be created.
        colors: tuple of two (2) color_likes, default: `("green", "red")`
            Color pair determining the color gradient with which the
            polar diagram will be plotted.

            Will be determined by the corresponding wind speeds.
        """
        wa = np.deg2rad(self.wind_angles)
        ws, wa = np.meshgrid(self.wind_speeds, wa)
        bsp = self.boat_speeds
        bsp, wa = bsp * np.cos(wa), bsp * np.sin(wa)

        plot_surface(ws, wa, bsp, ax, colors)

    def plot_color_gradient(
        self,
        ax=None,
        colors=("green", "red"),
        marker=None,
        ms=None,
        show_legend=False,
        **legend_kw,
    ):
        """Creates a 'wind speed vs. wind angle' color gradient plot
        of the polar diagram with respect to the corresponding boat speeds.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            Axes instance where the plot will be created.
        colors : tuple of two (2) color_likes, default: `("green", "red")`
            Color pair determining the color gradient with which the
            polar diagram will be plotted.

            Will be determined by the corresponding boat speed.
        marker : matplotlib.markers.Markerstyle or equivalent
            Markerstyle for the created scatter plot.
            Defaults to `"o"`.
        ms : array_like, optional
            Marker size in points**2.
        show_legend : bool, default: `False`
            Specifies whether a legend will be shown next
            to the plot.

            Legend will be a `matplotlib.colorbar.Colorbar` instance.
        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.
            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.
        """
        ws, wa = np.meshgrid(self.wind_speeds, self.wind_angles)
        ws = ws.ravel()
        wa = wa.ravel()
        bsp = self.boat_speeds.ravel()
        plot_color_gradient(
            ws, wa, bsp, ax, colors, marker, ms, show_legend, **legend_kw
        )

    def plot_convex_hull(
        self,
        ws=None,
        ax=None,
        colors=("green", "red"),
        show_legend=False,
        legend_kw=None,
        **plot_kw,
    ):
        """Computes the (separate) convex hull of one or more
        slices of the polar diagram and creates a polar plot of them.

        Parameters
        ----------
        ws : array_like, default: `self.wind_speeds`
            Slices of the polar diagram table, given as either:
            - a tuple of length 2 specifying an interval of considered
            wind speeds,
            - an iterable containing only elements of `self.wind_speeds`,
            - a single element of `self.wind_speeds`.

            The slices are then equal to the corresponding
            columns of the table together with `self.wind_angles`.
        ax : matplotlib.projections.polar.PolarAxes, optional
            Axes instance where the plot will be created.
        colors : sequence of color_likes or (ws, color_like) pairs, optional
            Specifies the colors to be used for the different slices.

            - If 2 colors are passed, slices will be plotted with a color
            gradient that is determined by the corresponding wind speed.
            - Otherwise the slices will be colored in turn with the specified
            colors or the color `"blue"`, if there are too few colors. The
            order is determined by the corresponding wind speeds.
            - Alternatively one can specify certain slices to be plotted in
            a color out of order by passing a sequence of `(ws, color)` pairs.

            Defaults to `("green", "red")`
        show_legend : bool, default: `False`
            Specifies whether a legend will be shown next to the plot.

            The type of legend depends on the color options.

            If plotted with a color gradient, a `matplotlib.colorbar.Colorbar`
            will be created, otherwise a `matplotlib.legend.Legend`.
        legend_kw : dict, optional
            Keyword arguments to change position and appearance of the colorbar
            or legend respectively.

            - If 2 colors are passed, a colorbar will be created.
            In this case see `matplotlib.colorbar.Colorbar` for possible
            keywords and their effect.
            - Otherwise, a legend will be created.
            In this case see `matplotlib.legend.Legend` for possible keywords
            and their effect.
        plot_kw : dict, optional
            Keyword arguments to change various appearances of the plot.

            See `matplotlib.axes.Axes.plot` for possible keywords and their
            effects.

        Raises
        ------
        PolarDiagramException
            If at least one element of `ws` is not in `self.wind_speeds`.
        PolarDiagramException
            If the given interval doesn't contain any slices of the
            polar diagram.
        """
        ws, wa, bsp = self.get_slices(ws)
        bsp = list(bsp.T)
        wa = [wa] * len(bsp)
        plot_convex_hull(
            ws,
            wa,
            bsp,
            ax,
            colors,
            show_legend,
            legend_kw,
            _lines=True,
            **plot_kw,
        )


class _Resolution(enum.Enum):
    WIND_SPEED = (np.arange(2, 42, 2), 40)
    WIND_ANGLE = (np.arange(0, 360, 5), 360)

    def __init__(self, standard_res, max_value):
        self.standard_res = standard_res
        self.max_value = max_value

    def set_resolution(self, res):
        if res is None:
            return self.standard_res

        if isinstance(res, Iterable):
            return self._custom_iterable_resolution(res)

        return self._custom_stepsize_resolution(res)

    def _custom_iterable_resolution(self, res):
        # NaN's and infinite values can cause problems later on
        res = np.asarray_chkfinite(res)

        if res.dtype == object:
            raise ValueError("`res` is not array_like")

        if not res.size or res.ndim != 1:
            raise ValueError("`res` has incorrect shape")

        if len(set(res)) != len(res):
            warnings.warn(
                "`res` contains duplicate data. "
                "This may lead to unwanted behaviour"
            )

        if self.name == _Resolution.WIND_SPEED.name:
            if np.any((res < 0)):
                raise ValueError("`res` contains negative entries")

        if self.name == _Resolution.WIND_ANGLE.name:
            res %= 360

        return res

    def _custom_stepsize_resolution(self, res):
        if res <= 0:
            raise ValueError("`res` is non-positive")

        if res > self.max_value:
            raise ValueError(
                f"Resolution stepsize ({res}) is bigger than the maximal"
                f" resolution value ({self.max_value})."
            )

        return np.arange(res, self.max_value, res)

    def normalize_wind(self, wind):
        wind = np.atleast_1d(wind)  # allow scalar inputs

        if self.name == _Resolution.WIND_SPEED.name:
            if np.any((wind < 0)):
                raise ValueError("`wind` is negative")

        if self.name == _Resolution.WIND_ANGLE.name:
            wind %= 360

        return set(wind)


def _incompatible_shapes(bsps, ws_resolution, wa_resolution):
    rows, cols = len(wa_resolution), len(ws_resolution)
    return bsps.shape != (rows, cols)


def _sort_table(ws_resolution, wa_resolution, bsps):
    wa_resolution, bsps = zip(
        *sorted(zip(wa_resolution, bsps), key=lambda x: x[0])
    )
    bsps = np.array(bsps, float)

    ws_resolution, bsps = zip(
        *sorted(zip(ws_resolution, bsps.T), key=lambda x: x[0])
    )

    return (
        np.array(ws_resolution, float),
        np.array(wa_resolution, float),
        np.array(bsps, float).T,
    )


def _delete_multiple_180_degree_occurrences(wa_resolution, bsps):
    mid = np.where(wa_resolution == 180)[0][0]
    wa_resolution = np.delete(wa_resolution, mid)
    bsps = np.row_stack((bsps[:mid, :], bsps[mid + 1 :, :]))

    return wa_resolution, bsps


def _delete_multiple_0_degree_occurrences(wa_resolution, bsps):
    return wa_resolution[:-1], bsps[:-1, :]



"""
This Module contains the abstract base class and inheriting classes for the
handling of weather information.
"""

import json
from abc import ABC, abstractmethod
from bisect import bisect_left
from datetime import datetime, timedelta

import numpy as np

try:
    import netCDF4 as nc

    installed_netcdf = True
except ModuleNotFoundError:
    installed_netcdf = False

try:
    import pandas as pd

    installed_pandas = True
except ModuleNotFoundError:
    installed_pandas = False

try:
    import meteostat

    installed_meteostat = True
except ModuleNotFoundError:
    installed_meteostat = False


class OutsideGridException(Exception):
    """Exception raised if point accessed in weather model lies
    outside the available grid"""


class WeatherModel(ABC):
    """
    Base class for handling and approximating weather data.
    How the weather data is organized and how the approximation is executed
    depends on the inheriting classes.

    Abstract Methods
    ----------------
    get_weather
    """

    @abstractmethod
    def get_weather(self, point):
        """Given a space-time point, uses the available weather model
        to calculate the weather at that point.

        Parameters
        ----------
        point: tuple of length 3
            Space-time point given as tuple of time, latitude
            and longitude.

        Returns
        -------
        weather : dict
            The weather data at the given point.

            If it is a grid point, the weather data is taken straight
            from the model, else it is interpolated as described above.
        """


class GriddedWeatherModel(WeatherModel):
    """Models a weather model as a 3-dimensional space-time grid
    where each space-time point has certain values of a given list
    of attributes. Points in between are approximated affinely.

    Parameters
    ----------
    data : array_like of shape (n, m, r, s)
        Weather data at different space-time grid points.

    times : list of length n
        Sorted list of time values of the space-time grid.

    lats : list of length m
        Sorted list of latitude values of the space-time grid.

    lons : list of length r
        Sorted list of longitude values of the space-time grid.

    attrs : list of length s
        List of different (scalar) attributes of weather.

    """

    def __init__(self, data, times, lats, lons, attrs):
        self._times = times
        self._lats = np.array(lats)
        self._lons = np.array(lons)
        self._attrs = attrs
        self._data = np.array(data)

    @property
    def grid(self):
        return self._times, self._lats.copy(), self._lons.copy()

    @property
    def data(self):
        return self._data.copy()

    @property
    def attrs(self):
        return self._attrs

    def get_weather(self, point):
        """Given a space-time point, uses the available weather model
        to calculate the weather at that point.

        If the point is not a grid point, the weather data will be
        interpolated via recursive affine interpolation
        of the up-to eight grid points which form a cuboid around the `points`.

        See also
        --------
        `WeatherModel.get_weather`
        """
        # check if given point lies in the grid
        val = _recursive_affine_interpolation(point, self.grid, self._get_data)

        return dict(zip(self._attrs, val))

    def _get_data(self, idxs):
        return self._data[idxs]

    @classmethod
    def from_meteostat(cls, lats, lons, start_time, end_time, keys):
        """
        Uses the `meteostat` module to fetch gridded weather data from the web.
        To use this method, you need to have the module `meteostat` installed
        and a connection to the internet.
        The time component of the resulting gridded data will have an hourly
        resolution.

        Parameters
        -----------

        lats : list of floats
            Sorted list of latitude values of the space-time grid.

        lons : list floats
            Sorted list of longitude values of the space-time grid.

        start_time : datetime.datetime
            The smallest time component of the grid.

        end_time : datetime.datetime
            (Approximately) the biggest time component of the grid.

        keys : list of str
            `meteostat` keys to be included in the weather model.

        Returns
        -------
        wm: `GriddedWeatherModel` as described above

        """
        if (not installed_meteostat) or (not installed_pandas):
            raise ImportError(
                """The modules `meteostat` and `pandas` are
                necessary to use `GriddedWeatherModel.from_meteostat`"""
            )

        # fetch data
        lat_datas = []
        for lat in lats:
            lon_data = []
            for lon in lons:
                loc = meteostat.Point(lat, lon, 0)
                data = meteostat.Hourly(loc, start_time, end_time).fetch()
                times = data.index.values
                np_data = data[keys].to_numpy()
                if np_data.shape[0] == 0:
                    np_data = []
                lon_data.append(np_data)
            max_len = max(len(d) for d in lon_data)
            lon_data = [
                np.zeros((max_len, len(keys))) if d == [] else d
                for d in lon_data
            ]
            lat_datas.append(np.stack(lon_data, axis=0))

        times = [pd.to_datetime(t) for t in times]
        data = np.stack(lat_datas, axis=0)
        data = np.transpose(np.stack(lat_datas, axis=0), axes=(2, 0, 1, 3))

        return cls(data, times, lats, lons, keys)

    def to_file(self, path):
        """
        Writes the data of the weather model to a `json` file such that it can
        be read via `from_file` method.

        Parameters
        ---------
        path : path like
            The path of the written file.
        """
        with open(path, "w", encoding="utf-8") as file:
            file.write(json.dumps(self, cls=_GriddedWeatherModelEncoder))

    @classmethod
    def from_file(cls, path):
        """
        Reads a gridded weather model from a compatible `json` file
        (for example those created via the `to_file` method)

        Parameters
        ---------
        path : path like
            The path of the file to be read.
        """
        with open(path, "r", encoding="utf-8") as file:
            data, times, *rest = json.loads(file.read())
            times = [datetime.strptime(t, "%d.%m.%Y:%X") for t in times]
            return cls(data, times, *rest)


class _GriddedWeatherModelEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, GriddedWeatherModel):
            data = o.data
            times, lats, lons = o.grid
            attrs = o.attrs
            return [data, times, lats, lons, attrs]
        if isinstance(o, np.ndarray):
            return o.tolist()
        if installed_pandas and isinstance(o, pd.Timestamp):
            return o.strftime("%d.%m.%Y:%X")
        if isinstance(o, datetime):
            return o.strftime("%d.%m.%Y:%X")
        raise TypeError(
            f"Object of type {type(o)} is not JSON serializable :("
        )


def _recursive_affine_interpolation(point, grid, get_data):
    fst = tuple(dimension[0] for dimension in grid)
    lst = tuple(dimension[-1] for dimension in grid)

    outside_left = [pt < left for pt, left in zip(point, fst)]
    outside_right = [pt > right for pt, right in zip(point, lst)]

    if any(outside_left) or any(outside_right):
        raise OutsideGridException(
            f"{point} is outside the grid. Weather data not available."
        )

    idxs = [
        bisect_left(grid_comp, comp) for grid_comp, comp in zip(grid, point)
    ]
    flags = [
        grid_pt[idx] == pt
        for grid_pt, idx, pt in zip(
            grid,
            idxs,
            point,
        )
    ]

    cuboid = [
        [idx - 1, idx] if not flag else [idx] for idx, flag in zip(idxs, flags)
    ]

    cuboid_vals = [
        [grid[dim][idx] for idx in c] for dim, c in enumerate(cuboid)
    ]

    def recursion(point_, completed=None):
        # get first entry which has not been computed yet
        if completed is None:
            completed = []

        dim = len(completed)

        if dim == len(point_):  # terminate
            return get_data(tuple(completed))

        comp = point_[dim]

        if comp in cuboid_vals[dim]:
            j = cuboid_vals[dim].index(point_[dim])
            return recursion(point_, completed + [j])

        idx0 = cuboid[dim][0]
        idx1 = cuboid[dim][1]

        val1 = grid[dim][idx1]
        val0 = grid[dim][idx0]

        lamb = (comp - val1) / (val0 - val1)

        data0 = point_[:dim] + [val0] + point_[dim + 1 :]
        data1 = point_[:dim] + [val1] + point_[dim + 1 :]

        term0 = recursion(data0, completed + [idx0])
        term1 = recursion(data1, completed + [idx1])

        return lamb * term0 + (1 - lamb) * term1

    return recursion(list(point))


class NetCDFWeatherModel(GriddedWeatherModel):
    """
    A weather model that uses gridded data from a NetCDF (.nc or .nc4) file.
    Uses the same interpolation method as `GriddedWeatherModel`.
    The module `netCDF4` has to be installed in order to use this class.
    The methods `from_file`, `to_file` and `from_meteostat` are not supported.

    Parameters
    ----------
    path: str
        Path to the NetCDF file to be used.

    aliases: dict with keys ["lat", "lon", "datetime"], optional
        Contains the aliases for latitude, longitude and the timestamp used in the NetCDF file.

        Defaults to `{"lat": "latitude", "lon": "longitude", "datetime": "time"}`.

    See also
    --------
    `GriddedWeatherModel`
    """

    def __init__(
        self,
        path,
        aliases=None,
        further_indices=None,
    ):
        if not installed_netcdf:
            raise ModuleNotFoundError(
                "Install netCDF4 to use NetCDFWeatherModel"
            )

        if aliases is None:
            aliases = {
                "lat": "latitude",
                "lon": "longitude",
                "datetime": "time",
            }
        if further_indices is None:
            further_indices = {}

        self._dataset = nc.Dataset(path)
        self._aliases = aliases
        self._further_indices = further_indices

        lats = self._dataset[aliases["lat"]]
        lons = self._dataset[aliases["lon"]]
        # check if in descending order and change order if necessary
        self._lats_flipped, self._lons_flipped = (
            lats[-1] < lats[0],
            lons[-1] < lons[0],
        )
        if self._lats_flipped:
            lats = np.flip(lats)
        if self._lons_flipped:
            lons = np.flip(lons)

        time = aliases["datetime"]
        plain_times = self._dataset[time][:]
        unit_str = self._dataset.variables[time].units
        if unit_str.startswith("hours since "):
            timestep = timedelta(hours=1)
            datetime_str = unit_str[12:]
        elif unit_str.startswith("seconds since "):
            timestep = timedelta(seconds=1)
            datetime_str = unit_str[14:]
        else:
            raise NotImplementedError(
                "NetCDFWeatherModel can not interpret time units"
                f" '{unit_str}'."
            )
        fmts = ["%Y-%m-%d %H:%M", "%Y-%m-%d %H:%M:%S"]
        starting_time = None
        for fmt in fmts:
            try:
                starting_time = datetime.strptime(datetime_str, fmt)
                break
            except ValueError:
                pass
        if starting_time is None:
            raise NotImplementedError(
                "NetCDFWeatherModel does not support the time format of"
                f" `{datetime_str}`. Use one of {fmts} instead."
            )

        times = [
            starting_time + int(plain_time) * timestep
            for plain_time in plain_times
        ]

        attrs = [
            var
            for var in self._dataset.variables
            if var not in aliases.values()
        ]

        super().__init__(None, times, lats, lons, attrs)

    def _get_data(self, idxs):
        times, lats, lons = self.grid
        time_idx = idxs[0]
        lat_idx = len(lats) - idxs[1] - 1 if self._lats_flipped else idxs[1]
        lon_idx = len(lons) - idxs[2] - 1 if self._lons_flipped else idxs[2]

        idxs = (time_idx, lat_idx, lon_idx)

        attr_list = []
        for attr in self._attrs:
            try:
                coords = self._get_dimensions(attr)
            except AttributeError:
                attr_list.append(self._dataset[attr][idxs])
                continue
            new_idxs = []
            for coord in coords:
                if coord == self._aliases["lat"]:
                    new_idxs.append(lat_idx)
                elif coord == self._aliases["lon"]:
                    new_idxs.append(lon_idx)
                elif coord == self._aliases["datetime"]:
                    new_idxs.append(time_idx)
                elif coord in self._further_indices:
                    new_idxs.append(self._further_indices[coord])
                else:
                    raise NotImplementedError(
                        "For the handling of variables other than 'latitude',"
                        f" 'longitude' and 'time' as '{coord}' initialize the"
                        " NetCDFWeatherModel with "
                        f"'further_indices={{'{coord}': a}}'"
                    )
            attr_list.append(self._dataset[attr][tuple(new_idxs)])

        return np.asarray(attr_list)

    def _get_dimensions(self, attr):
        dimensions = str(self._dataset[attr]).split("\n")[1].split(" ")[1:]
        dimensions = [
            c.split("(")[1].strip(",)") if "(" in c else c.strip(",)")
            for c in dimensions
        ]
        return dimensions


class MultiWeatherModel(WeatherModel):
    """
    Weather model that manages multiple weather models at once and combines their output.

    Parameters
    ----------
    *args
        An arbitrary number of weather models.

    exception_sensitive: bool, optional
        If `False` any `ValueError`, `TypeError`, `KeyError`, `NotImplementedError` or `OutsideGridException`
        raised by some submodel is ignored. If `True` these exceptions are not ignored.

        Defaults to `False`.
    """

    def __init__(self, *args, exception_sensitive=False):
        self._models = args
        self._exception_sensitive = exception_sensitive

    def get_weather(self, point):
        """
        Evaluates all submodels and combines the results.

        See also
        --------
        `WeatherModel.get_weather`
        """
        weather_data = {}
        for model in self._models:
            try:
                weather_data.update(model.get_weather(point))
            except (
                ValueError,
                TypeError,
                KeyError,
                OutsideGridException,
            ) as e:
                if self._exception_sensitive:
                    raise e
        return weather_data



"""
Contains the abstract base class `PipelineExtension` and some ready-to-use implementations of it.
Pipeline extensions should take preprocessed data and use it to create polar diagrams.
"""

import warnings
from abc import ABC, abstractmethod

import numpy as np

import hrosailing.pipelinecomponents as pc
import hrosailing.polardiagram as pol
from hrosailing.pipelinecomponents._utils import ComponentWithStatistics
from hrosailing.pipelinecomponents.modelfunctions import (
    ws_s_wa_gauss_and_square,
)
from hrosailing.polardiagram._polardiagramtable import _Resolution


class PipelineExtension(ComponentWithStatistics, ABC):
    """Base class for all pipeline extensions.

    Abstract Methods
    ----------------
    process(weighted_points)
    """

    @abstractmethod
    def process(self, weighted_points):
        """This method, given an instance of WeightedPoints, should
        return a polar diagram object, which represents the trends
        and data contained in the WeightedPoints instance.

        Parameters
        -----------
        weighted_points : WeightedPoints
            Preprocessed data from which to create the polar diagram.
            `weighted_points.data` has to be an `np.ndarray` of shape (n, 3) of dtype `float` containing
            'TWS', 'TWA', 'BSP' values columnwise.

        Returns
        -------
        polar_diagram : PolarDiagramTable
            A polar diagram that should represent the trends
            captured in the raw data.
        """


class TableExtension(PipelineExtension):
    """Pipeline extension to produce `PolarDiagramTable` instances
    from preprocessed data.

    Parameters
    ----------
    wind_resolution : tuple of two array_likes or scalars, or str, optional
        Wind speed and angle resolution to be used in the final table.
        Can be given as:

        - a tuple of two `array_likes` with scalar entries, that
        will be used as the resolution,
        - a tuple of two `scalars`, which will be used as
        step-sizes for the resolutions,
        - the str `"auto"`, which will result in a resolution, that is
        somewhat fitted to the data.
        - the str `"min_max"`, which will result in a resolution defined by the
        minimum and maximum of the data

        Defaults to `"min_max"`.

    neighbourhood : Neighbourhood, optional
        Determines the neighbourhood around a point from which to draw
        the data points used in the interpolation of that point.

        Defaults to `Ball(radius=1)`.

    interpolator : Interpolator, optional
        Determines which interpolation method is used.

        Defaults to `ArithmeticMeanInterpolator(params=(50,))`.
    """

    def __init__(
        self,
        wind_resolution="min_max",
        neighbourhood=pc.Ball(radius=1),
        interpolator=pc.ArithmeticMeanInterpolator(params=(50,)),
    ):
        super().__init__()
        self.wind_resolution = wind_resolution
        self.neighbourhood = neighbourhood
        self.interpolator = interpolator

    def process(self, weighted_points):
        """Creates a `PolarDiagramTable` instance from preprocessed data,
        by first determining a wind speed / wind angle grid, using
        `self.w_res`, and then interpolating the boat speed values at the
        grid points according to the interpolation method of
        `self.interpolator`, which only takes in consideration the data points
        which lie in a neighbourhood, determined by `self.neighbourhood`,
        around each grid point.

        Returns
        -------
        polar_diagram : PolarDiagramTable

        See also
        --------
        `PipelineExtension.process`
        """
        ws_resolution, wa_resolution = self._determine_table_size(
            weighted_points.data
        )
        ws, wa = np.meshgrid(ws_resolution, wa_resolution)
        grid_points = np.column_stack((ws.ravel(), wa.ravel()))

        interpolated_points = _interpolate_points(
            grid_points, weighted_points, self.neighbourhood, self.interpolator
        )
        bsps = _extract_boat_speed(
            interpolated_points, len(wa_resolution), len(ws_resolution)
        )

        return pol.PolarDiagramTable(ws_resolution, wa_resolution, bsps)

    def _determine_table_size(self, points):
        if self.wind_resolution == "auto":
            return _automatically_determined_resolution(points)

        if self.wind_resolution == "min_max":
            return _min_max_resolution(points)
        if self.wind_resolution is None:
            self.wind_resolution = (None, None)

        ws_resolution, wa_resolution = self.wind_resolution
        return (
            _Resolution.WIND_SPEED.set_resolution(ws_resolution),
            _Resolution.WIND_ANGLE.set_resolution(wa_resolution),
        )


def _automatically_determined_resolution(points):
    ws_resolution = _extract_wind(points[:, 0], 2, 100)
    wa_resolution = _extract_wind(points[:, 1], 5, 30)

    return ws_resolution, wa_resolution


def _min_max_resolution(points):
    ws_min = round(points[:, 0].min())
    ws_max = round(points[:, 0].max())
    ws_resolution = list(range(ws_min, ws_max + 1, 2))
    wa_resolution = list(range(5, 356, 5))

    return ws_resolution, wa_resolution


def _extract_wind(points, n, threshhold):
    w_max = round(points.max())
    w_min = round(points.min())
    w_start = (w_min // n + 1) * n
    w_end = (w_max // n) * n
    res = [w_max, w_min]

    for w in range(w_start, w_end + n, n):
        if w == w_start:
            mask = np.logical_and(w >= points, points >= w_min)
        elif w == w_end:
            mask = np.logical_and(w_max >= points, points >= w)
        else:
            mask = np.logical_and(w >= points, points >= w - n)

        if len(points[mask]) >= threshhold:
            res.append(w)

    return res


def _extract_boat_speed(interpolated_points, rows, cols):
    bsps = interpolated_points[:, 2]
    return bsps.reshape(rows, cols)


class CurveExtension(PipelineExtension):
    """Pipeline extension to produce `PolarDiagramCurve` instances
    from preprocessed data.

    Parameters
    ----------
    regressor : Regressor, optional
        Determines which regression method and model function is to be used,
        to represent the data.

        The model function will also be passed to `PolarDiagramCurve`.

        Defaults to `ODRegressor(
            model_func=ws_s_wa_gauss_and_square,
            init_values=(0.2, 0.2, 10, 0.001, 0.3, 110, 2000, 0.3, 250, 2000)
        )`.

    radians : bool, optional
        Determines if the model function used to represent the data takes
        the wind angles to be in radians or degrees.

        If `True`, will convert the wind angles of the data points to
        radians.

        Defaults to `False`.
    """

    def __init__(
        self,
        regressor=pc.ODRegressor(
            model_func=ws_s_wa_gauss_and_square,
            init_values=(0.2, 0.2, 10, 0.001, 0.3, 110, 2000, 0.3, 250, 2000),
        ),
        radians=False,
    ):
        super().__init__()
        self.regressor = regressor
        self.radians = radians

    def process(self, weighted_points):
        """Creates a `PolarDiagramCurve` instance from preprocessed data,
        by fitting a given function to said data, using a regression
        method determined by `self.regressor`.

        See also
        --------
        `PipelineExtension.process`
        """
        if self._use_radians():
            _convert_angles_to_radians(weighted_points)

        self.regressor.fit(weighted_points.data)

        return pol.PolarDiagramCurve(
            self.regressor.model_func,
            *self.regressor.optimal_params,
            radians=self.radians,
        )

    def _use_radians(self):
        return self.radians


def _convert_angles_to_radians(weighted_points):
    weighted_points.data[:, 1] = np.deg2rad(weighted_points.data[:, 1])


class PointcloudExtension(PipelineExtension):
    """Pipeline extension to produce `PolarDiagramPointcloud` instances
    from preprocessed data.

    Parameters
    ----------
    sampler : Sampler, optional
        Determines the number of points in the resulting point cloud
        and the method used to sample the preprocessed data and represent
        the trends captured in them.

        Defaults to `UniformRandomSampler(2000)`.

    neighbourhood : Neighbourhood, optional
        Determines the neighbourhood around a point from which to draw
        the data points used in the interpolation of that point.

        Defaults to `Ball(radius=1)`.

    interpolator : Interpolator, optional
        Determines which interpolation method is used.

        Defaults to `ArithmeticMeanInterpolator(params=(50,))`.
    """

    def __init__(
        self,
        sampler=pc.UniformRandomSampler(2000),
        neighbourhood=pc.Ball(radius=1),
        interpolator=pc.ArithmeticMeanInterpolator(params=(50,)),
    ):
        super().__init__()
        self.sampler = sampler
        self.neighbourhood = neighbourhood
        self.interpolator = interpolator

    def process(self, weighted_points):
        """Creates a `PolarDiagramPointcloud` instance from preprocessed data,
        first creating a set number of points by sampling the wind speed,
        wind angle, space of the data points and capturing the underlying
        trends using `self.sampler` and then interpolating the boat speed
        values at the sampled points according to the interpolation method of
        `self.interpolator`, which only takes in consideration the data points
        which lie in a neighbourhood, determined by `self.neighbourhood`,
        around each sampled point.

        Returns
        -------
        pd: PolarDiagramPointcloud

        See also
        --------
        `PipelineExtension.process`
        """
        sample_points = self.sampler.sample(weighted_points.data)
        interpolated_points = _interpolate_points(
            sample_points,
            weighted_points,
            self.neighbourhood,
            self.interpolator,
        )

        return pol.PolarDiagramPointcloud(interpolated_points)


class InterpolationWarning(Warning):
    """Warning raised if neighbourhood is too small for
    successful interpolation
    """


def _interpolate_points(
    interpolating_points, weighted_points, neighbourhood, interpolator
):
    interpolated_points = [
        _interpolate_point(point, weighted_points, neighbourhood, interpolator)
        for point in interpolating_points
    ]

    return np.array(interpolated_points)


def _interpolate_point(point, weighted_points, neighbourhood, interpolator):
    considered = neighbourhood.is_contained_in(
        weighted_points.data[:, :2] - point
    )

    if _neighbourhood_too_small(considered):
        warnings.warn(
            (
                "Neighbourhood possibly to `small`, or"
                "chosen resolution not fitting for data. "
                "Interpolation will not lead to complete results"
            ),
            category=InterpolationWarning,
        )
        return np.concatenate([point, [0]])

    interpolated_value = interpolator.interpolate(
        weighted_points[considered], point
    )

    return np.concatenate([point, [interpolated_value]])


def _neighbourhood_too_small(considered_points):
    return not np.any(considered_points)



"""Contains various helper functions for the `plot_*`-methods()."""

# pylint: disable=missing-function-docstring


import itertools

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.cm import ScalarMappable
from matplotlib.colors import (
    LinearSegmentedColormap,
    Normalize,
    is_color_like,
    to_rgb,
)
from matplotlib.lines import Line2D
from scipy.spatial import ConvexHull


def plot_polar(
    ws, wa, bsp, ax, colors, show_legend, legend_kw, _lines, **plot_kw
):
    if ax is None:
        ax = _get_new_axis("polar")
    _set_polar_axis(ax)

    _check_plot_kw(plot_kw, _lines)

    _plot(ws, wa, bsp, ax, colors, show_legend, legend_kw, **plot_kw)


def _get_new_axis(kind):
    return plt.axes(projection=kind)


def _set_polar_axis(ax):
    ax.set_theta_zero_location("N")
    ax.set_theta_direction("clockwise")


def _check_plot_kw(plot_kw, lines=True):
    ls = plot_kw.pop("linestyle", None) or plot_kw.pop("ls", None)
    if ls is None:
        plot_kw["ls"] = "-" if lines else ""
    else:
        plot_kw["ls"] = ls

    if plot_kw.get("marker", None) is None and not lines:
        plot_kw["marker"] = "o"


def _plot(ws, wa, bsp, ax, colors, show_legend, legend_kw, **plot_kw):
    _configure_colors(ax, ws, colors)

    if _only_one_color(colors):
        colors = [colors] * len(ws)

    if show_legend:
        _show_legend(ax, ws, colors, "True Wind Speed", legend_kw)

    for x, y in zip(wa, bsp):
        ax.plot(x, y, **plot_kw)


def _configure_colors(ax, ws, colors):
    if _only_one_color(colors):
        ax.set_prop_cycle("color", [colors])
        return

    if _more_colors_than_plots(ws, colors) or _no_color_gradient(colors):
        _set_color_cycle(ax, ws, colors)
        return

    _set_color_gradient(ax, ws, colors)


def _only_one_color(colors):
    return is_color_like(colors)


def _more_colors_than_plots(ws, colors):
    return len(ws) <= len(colors)


def _no_color_gradient(colors):
    all_color_format = all(_has_color_format(c) for c in colors)
    return len(colors) != 2 or not all_color_format


def _has_color_format(obj):
    if isinstance(obj, str):
        return True
    if len(obj) in [3, 4]:
        return True
    return False


def _set_color_cycle(ax, ws, colors):
    color_cycle = ["blue"] * len(ws)
    _configure_color_cycle(color_cycle, colors, ws)

    ax.set_prop_cycle("color", color_cycle)


def _configure_color_cycle(color_cycle, colors, ws):
    if isinstance(colors[0], tuple):
        for w, color in colors:
            i = list(ws).index(w)
            color_cycle[i] = color

        return

    colors = itertools.islice(colors, len(color_cycle))

    for i, color in enumerate(colors):
        color_cycle[i] = color


def _set_color_gradient(ax, ws, colors):
    color_gradient = _determine_color_gradient(colors, ws)
    ax.set_prop_cycle("color", color_gradient)


def _determine_color_gradient(colors, gradient):
    gradient_coeffs = _get_gradient_coefficients(gradient)
    color_gradient = _determine_colors_from_coefficients(
        gradient_coeffs, colors
    )
    return color_gradient


def _get_gradient_coefficients(gradient):
    min_gradient = min(gradient)
    max_gradient = max(gradient)

    return [
        (grad - min_gradient) / (max_gradient - min_gradient)
        for grad in gradient
    ]


def _determine_colors_from_coefficients(coefficients, colors):
    min_color = np.array(to_rgb(colors[0]))
    max_color = np.array(to_rgb(colors[1]))

    return [
        (1 - coeff) * min_color + coeff * max_color for coeff in coefficients
    ]


def _show_legend(ax, ws, colors, label, legend_kw):
    if legend_kw is None:
        legend_kw = {}

    _configure_legend(ax, ws, colors, label, **legend_kw)


def _configure_legend(ax, ws, colors, label, **legend_kw):
    if _plot_with_color_gradient(ws, colors):
        _set_colormap(ws, colors, ax, label, **legend_kw)
        return

    if isinstance(colors[0], tuple) and not is_color_like(colors[0]):
        _set_legend_without_wind_speeds(ax, colors, legend_kw)
        return

    _set_legend_with_wind_speeds(ax, colors, ws, legend_kw)


def _plot_with_color_gradient(ws, colors):
    return not _no_color_gradient(colors) and len(ws) > len(colors) == 2


def _set_colormap(ws, colors, ax, label, **legend_kw):
    color_map = _create_color_map(colors)

    label_kw, legend_kw = _extract_possible_text_kw(legend_kw)
    plt.colorbar(
        ScalarMappable(
            norm=Normalize(vmin=min(ws), vmax=max(ws)), cmap=color_map
        ),
        ax=ax,
        **legend_kw,
    ).set_label(label, **label_kw)


def _extract_possible_text_kw(legend_kw):
    return {}, legend_kw


def _set_legend_without_wind_speeds(ax, colors, legend_kw):
    ax.legend(
        handles=[
            Line2D([0], [0], color=color, lw=1, label=f"TWS {ws}")
            for (ws, color) in colors
        ],
        **legend_kw,
    )


def _set_legend_with_wind_speeds(ax, colors, ws, legend_kw):
    slices = zip(ws, colors)

    ax.legend(
        handles=[
            Line2D([0], [0], color=color, lw=1, label=f"TWS {ws}")
            for (ws, color) in slices
        ],
        **legend_kw,
    )


def plot_flat(
    ws, wa, bsp, ax, colors, show_legend, legend_kw, _lines, **plot_kw
):
    if ax is None:
        ax = _get_new_axis("rectilinear")

    _check_plot_kw(plot_kw, _lines)

    _plot(ws, wa, bsp, ax, colors, show_legend, legend_kw, **plot_kw)


def plot_color_gradient(
    ws, wa, bsp, ax, colors, marker, ms, show_legend, **legend_kw
):
    if ax is None:
        ax = _get_new_axis("rectilinear")

    if show_legend:
        _show_legend(ax, bsp, colors, "Boat Speed", legend_kw)

    color_gradient = _determine_color_gradient(colors, bsp)

    ax.scatter(ws, wa, s=ms, marker=marker, c=color_gradient)


def plot3d(ws, wa, bsp, ax, colors, **plot_kw):
    if ax is None:
        ax = _get_new_axis("3d")

    _set_3d_axis_labels(ax)
    _remove_3d_tick_labels_for_polar_coordinates(ax)

    color_map = _create_color_map(colors)

    ax.scatter(ws, wa, bsp, c=ws, cmap=color_map, **plot_kw)


def _set_3d_axis_labels(ax):
    ax.set_xlabel("TWS")
    ax.set_ylabel("Polar plane: TWA / BSP ")


def _remove_3d_tick_labels_for_polar_coordinates(ax):
    ax.yaxis.set_ticklabels([])
    ax.zaxis.set_ticklabels([])


def _create_color_map(colors):
    return LinearSegmentedColormap.from_list("cmap", list(colors))


def plot_surface(ws, wa, bsp, ax, colors):
    if ax is None:
        ax = _get_new_axis("3d")

    _set_3d_axis_labels(ax)
    _remove_3d_tick_labels_for_polar_coordinates(ax)

    color_map = _create_color_map(colors)
    face_colors = _determine_face_colors(color_map, ws)

    ax.plot_surface(ws, wa, bsp, facecolors=face_colors)


def _determine_face_colors(color_map, ws):
    return color_map((ws - ws.min()) / float((ws - ws.min()).max()))


def plot_convex_hull(
    ws, wa, bsp, ax, colors, show_legend, legend_kw, _lines, **plot_kw
):
    if ax is None:
        ax = _get_new_axis("polar")
    _set_polar_axis(ax)

    _check_plot_kw(plot_kw, _lines)

    wa, bsp = _convex_hull(zip(wa, bsp))

    _plot(ws, wa, bsp, ax, colors, show_legend, legend_kw, **plot_kw)


def _convex_hull(slices):
    xs, ys = [], []
    for wa, bsp in slices:
        wa = np.asarray(wa)
        bsp = np.asarray(bsp)

        # convex hull is line between the two points
        # or is equal to one point
        if len(wa) < 3:
            xs.append(wa)
            ys.append(bsp)
            continue

        conv = _convex_hull_in_polar_coordinates(wa, bsp)
        vert = conv.vertices
        x, y = zip(
            *([(wa[i], bsp[i]) for i in vert] + [(wa[vert[0]], bsp[vert[0]])])
        )
        xs.append(list(x))
        ys.append(list(y))

    return xs, ys


def _convex_hull_in_polar_coordinates(wa, bsp):
    polar_points = np.column_stack((bsp * np.cos(wa), bsp * np.sin(wa)))
    return ConvexHull(polar_points)


def plot_convex_hull_multisails(
    ws, wa, bsp, members, ax, colors, show_legend, legend_kw, **plot_kw
):
    if ax is None:
        ax = _get_new_axis("polar")

    _set_polar_axis(ax)

    _check_plot_kw(plot_kw)

    xs, ys, members = _get_convex_hull_multisails(ws, wa, bsp, members)

    if colors is None:
        colors = plot_kw.pop("color", None) or plot_kw.pop("c", None) or []
    colors = dict(colors)
    _set_colors_multisails(ax, members, colors)

    if legend_kw is None:
        legend_kw = {}
    if show_legend:
        _set_legend_multisails(ax, colors, **legend_kw)

    for x, y in zip(list(xs), list(ys)):
        ax.plot(x, y, **plot_kw)


def _get_convex_hull_multisails(ws, wa, bsp, members):
    xs = []
    ys = []
    membs = []
    for s, w, b in zip(ws, wa, bsp):
        w = np.asarray(w)
        b = np.asarray(b)
        conv = _convex_hull_in_polar_coordinates(w, b)
        vert = sorted(conv.vertices)

        x, y, memb = zip(
            *(
                [(w[i], b[i], members[i]) for i in vert]
                + [(w[vert[0]], b[vert[0]], members[vert[0]])]
            )
        )
        x = list(x)
        y = list(y)
        memb = list(memb)

        for i in range(len(vert)):
            xs.append(x[i : i + 2])
            ys.append(y[i : i + 2])
            membs.append(memb[i : i + 2] + [s])

    return xs, ys, membs


def _set_colors_multisails(ax, members, colors):
    colorlist = []

    for member in members:
        # check if edge belongs to one or two sails
        # If it belongs to one sail, color it in that sails color
        # else color it in neutral color
        if len(set(member[:2])) == 1:
            color = colors.get(member[0], "blue")
        else:
            color = colors.get("neutral", "gray")

        if is_color_like(color):
            colorlist.append(color)
            continue

        color = dict(color)
        colorlist.append(color.get(member[2], "blue"))

    ax.set_prop_cycle("color", colorlist)


def _set_legend_multisails(ax, colors, **legend_kw):
    handles = []
    for key in colors:
        color = colors.get(key, "blue")

        if is_color_like(color):
            legend = Line2D([0], [0], color=color, lw=1, label=key)
            handles.append(legend)
            continue

        color = dict(color)
        legends = [
            Line2D(
                [0],
                [0],
                color=color.get(ws, "blue"),
                lw=1,
                label=f"{key} at TWS {ws}",
            )
            for ws in color
        ]
        handles.extend(legends)

    if "neutral" not in colors:
        legend = Line2D([0], [0], color="gray", lw=1, label="neutral")
        handles.append(legend)

    ax.legend(handles=handles, **legend_kw)



"""
Defines the `InfluenceModel` abstract base class that can be used
to create custom influence models.

Subclasses of `InfluenceModel` can be used with

- the `PolarPipeline` class in the `hrosailing.pipeline` module,
- various functions in the `hrosailing.cruising` module.
"""

from abc import ABC, abstractmethod

import numpy as np

from hrosailing.pipelinecomponents._utils import ComponentWithStatistics
from hrosailing.wind import convert_apparent_wind_to_true

from ._utils import data_dict_to_numpy


class InfluenceException(Exception):
    """Raised when removing or adding influence does not work."""


class InfluenceModel(ComponentWithStatistics, ABC):
    """Base class for all influence model classes.


    Abstract Methods
    ----------------
    remove_influence(data)

    add_influence(pd, influence_data)

    fit(training_data)
    """

    @abstractmethod
    def remove_influence(self, data):
        """This method should be used to create a `numpy.ndarray`
        output from a given `Data` object where the columns correspond to wind speed, wind angle and
        boat speed respectively.

        Parameters
        ----------
        data : Data
            Should contain at least keys for wind speed, wind angle
            and either speed over ground, speed over water or boat speed.

        Returns
        -------
        out : numpy.ndarray
        """

    @abstractmethod
    def add_influence(self, pd, influence_data):
        """This method should be used, given a polar diagram and a
        dictionary, to obtain a modified boat speed of that given
        in the polar diagram, based on the influences presented in
        the given dictionary, such as wave height, underlying currents etc.

        Parameters
        ----------
        pd : PolarDiagram

        influence_data : Data or dict or list of Data or list of dict
            Further influences to be considered.

        Returns
        -------
        speeds : float or list of floats
            The boat speed if `influence_data` contained values,
            a list of respective boat speeds if `influence_data` contained
            lists.
        """

    def fit(self, training_data):
        """
        This method can be overwritten in order
        to fit parameters of the influence
        model to given training data.

        Parameters
        ----------
        training_data : Data
        """


class IdentityInfluenceModel(InfluenceModel):
    """An influence model which ignores most influences and just calculates
    the true wind if only the apparent wind is given.
    IF 'BSP' is not provided by the data, 'SOG' is used instead."""

    def remove_influence(self, data: dict):
        """
        Ignores most influences as described above.

        Parameters
        ----------
        data : dict
            Data dictionary, must either provide 'BSP' or 'SOG' key as well as
            either the keys 'TWS', 'TWA' or 'AWS', 'AWA'.

        Returns
        -------
        (n,3)-array, {}

        See also
        --------
        `InfluenceModel.remove_influence`
        """
        return _get_true_wind_data(data)

    def add_influence(self, pd, influence_data: dict):
        """
        Ignores most influences and strictly calculates the boat speed using the
        polar diagram.

        Parameters
        ----------
        pd : PolarDiagram

        influence_data : dict
            Either a dictionary of lists or a dictionary of values containing
            one or more sets of influence data.
            At least the keys 'TWS' and 'TWA' should be provided.

        See also
        --------
        `InfluenceModel.add_influence`
        """
        if isinstance(influence_data["TWS"], list):
            wind = zip(influence_data["TWS"], influence_data["TWA"])
            speed = [pd(ws, wa) for ws, wa in wind]
        else:
            ws, wa = influence_data["TWS"], influence_data["TWA"]
            speed = pd(ws, wa)
        return speed

    def fit(self, training_data: dict):
        """
        Does nothing.

        See also
        --------
        `InfluenceModel.fit`
        """


class WindAngleCorrectingInfluenceModel(InfluenceModel):
    """
    An influence model which corrects a structural measurement error in 'TWA'.

    Parameters
    ----------
    wa_shift: int or float, optional
        Difference between real wind angle and measured wind angle (correction value).

        Defaults to 0.

    interval_size: int or float, optional
        Interval size used in the fitting method.

        Defaults to 30.
    """

    def __init__(self, interval_size=30, wa_shift=0):
        super().__init__()
        self._wa_shift = wa_shift
        self._interval_size = interval_size

    def remove_influence(self, data):
        """
        Removes the correction value from the data.

        See also
        -------
        `InfluenceModel.remove_influence`
        """
        wind_data = _get_true_wind_data(data)
        wa = wind_data[:, 1]
        wind_data[:, 1] = (wa - self._wa_shift) % 360
        return wind_data

    def add_influence(self, pd, influence_data):
        """
        Adds the correction value to the data.

        See also
        -------
        `InfluenceModel.add_influence`

        """
        if isinstance(influence_data["TWS"], (list, np.ndarray)):
            wind = zip(influence_data["TWS"], influence_data["TWA"])
            speed = [pd(ws, (wa + self._wa_shift) % 360) for ws, wa in wind]
        else:
            ws, wa = influence_data["TWS"], influence_data["TWA"]
            speed = pd(ws, wa)
        return speed

    def fit(self, training_data):
        """
        The wind angle with the lowest density of measured wind angles is assumed to be the
        actual zero. The data density is computed using gauss kernel functions
        :math:`e^{\frac{wa - wa'}{l}}`, where :math:`l` can be seen as the size of
        an interval on which the gauss kernel is stretched.

        Parameter
        ---------
        training_data: Data containing key "TWA"

        See also
        --------
        `InfluenceModel.fit`
        """
        wind_angles = training_data["TWA"]
        wind_speeds = training_data["TWS"]
        sample = np.linspace(0, 359, 10 * 360)
        counts = [
            sum(
                other_ws
                * np.exp(
                    -(((abs(wa - other_wa) % 360) / self._interval_size) ** 2)
                )
                for other_wa, other_ws in zip(wind_angles, wind_speeds)
            )
            for wa in sample
        ]
        min_angle, _ = min(zip(sample, counts), key=lambda x: x[1])
        self._wa_shift = min_angle
        self.set_statistics(wa_shift=min_angle)


def _get_true_wind_data(data: dict):
    speed = "BSP" if "BSP" in data else "SOG"
    if "TWA" in data and "TWS" in data:
        if isinstance(data["TWS"], list):
            return data_dict_to_numpy(data, ["TWS", "TWA", speed])

        return np.array([data["TWS"], data["TWA"], data[speed]])

    if "AWA" in data and "AWS" in data:
        apparent_data = data_dict_to_numpy(data, ["AWS", "AWA", speed])
        return convert_apparent_wind_to_true(apparent_data)

    raise InfluenceException(
        "No sufficient wind data is given in order to apply influence"
        "model. Either give 'AWA' and 'AWS' or 'TWA' and 'TWS'"
    )



"""
Classes used for modular modelling of different sampling methods.

Defines the `Sampler` abstract base class that can be used to create
custom sampling methods.

Subclasses of `Sampler` can be used with the `PointcloudExtension` class
in the `hrosailing.pipeline` module.
"""


from abc import ABC, abstractmethod

import numpy as np
from scipy.spatial import ConvexHull


class SamplerInitializationException(Exception):
    """Exception raised if an error occurs during
    initialization of a `Sampler` object.
    """


class Sampler(ABC):
    """Base class for all sampler classes.


    Abstract Methods
    ----------------
    sample(self, pts)
    """

    @abstractmethod
    def sample(self, pts):
        """This method should be used, given certain points, to determine a
        constant number of sample points that are more or less representative of the trend of the given points.

        Parameters
        ----------
        pts : numpy.ndarray
            The given data points in a row-wise fashion. Most commonly this should be an (n, 2) array.

        Returns
        ----------
        sample : numpy.ndarray
            A sample representing the data points.
        """


class UniformRandomSampler(Sampler):
    """A sampler that produces a number of uniformly distributed samples,
    which all lie in the convex hull of certain given points.

    Parameters
    ----------
    n_samples : positive int
        Amount of samples that will be produced by the sampler.

    Raises
    ------
    SamplerInitializationException
        If `n_samples` is non-positive.
    """

    def __init__(self, n_samples):
        if n_samples <= 0:
            raise SamplerInitializationException("`n_samples` is non-positive")

        self._n_samples = n_samples

    def sample(self, pts):
        """Produces samples according to the procedure described above.

        Parameters
        ----------
        pts : array_like of shape (n, 2)
            Points in whose convex hull the produced samples will lie.

        Returns
        -------
        samples : numpy.ndarray of shape (n_samples, 2)
            Samples produced by the procedure described above.
        """
        rng = np.random.default_rng()
        proj_pts = pts[:, :2]
        ineqs = ConvexHull(proj_pts).equations
        samples = []
        ws_bound, wa_bound = _create_bounds(pts)
        while len(samples) < self._n_samples:
            ws = rng.uniform(
                ws_bound[0], ws_bound[1], self._n_samples - len(samples)
            )
            wa = rng.uniform(
                wa_bound[0], wa_bound[1], self._n_samples - len(samples)
            )
            wind = np.column_stack((ws, wa))
            mask = np.all((ineqs[:, :2] @ wind.T).T <= -ineqs[:, 2], axis=1)
            samples.extend(wind[mask])

        return np.asarray(samples)


def _create_bounds(pts):
    ws_bound = (np.amin(pts[:, 0]), np.amax(pts[:, 0]))
    wa_bound = (np.amin(pts[:, 1]), np.amax(pts[:, 1]))
    return ws_bound, wa_bound


class FibonacciSampler(Sampler):
    """A sampler that produces sample points on a moved and scaled version
    of the spiral (sqrt(x) * cos(x), sqrt(x) * sin(x)), such that the angles
    are distributed equidistantly by the inverse golden ratio.

    The sample points all lie in the smallest enclosing circle
    of given data points.

    Inspired by Álvaro Gonzzález - "Measurement of areas on a sphere using
    Fibonacci and latitude–longitude lattices".

    Parameters
    ----------
    n_samples : positive int
        Amount of samples that will be produced by the sampler.

    Raises
    ------
    SamplerInitializationException
        If `n_samples` is non-positive.
    """

    def __init__(self, n_samples):
        if n_samples <= 0:
            raise SamplerInitializationException("`n_samples` is non-positive")

        self._n_samples = n_samples

    def sample(self, pts):
        """Produces samples according to the procedure described above.

        Parameters
        ----------
        pts : array_like of shape (n, 2)
            Points in whose convex hull the produced samples will lie.

        Returns
        -------
        samples : numpy.ndarray of shape (n_samples, 2)
            Samples produced by the procedure described above.
        """
        # calculate smallest circle containing pts
        midpoint, r = _make_circle(pts)

        # calculate an upper bound to the number of points needed for the
        # spiral to fill the convex hull with self._n_samples points
        ch = ConvexHull(pts)
        vol = ch.volume
        ineqs = ch.equations
        ub_n_samples = int(np.pi * r**2 * self._n_samples / vol) + 10

        # create big fibonacci spiral
        golden_ratio = (1 + np.sqrt(5)) / 2
        i = np.arange(1, ub_n_samples)
        beta = 2 * np.pi * i * golden_ratio ** (-1)
        radius = np.sqrt(i / ub_n_samples)
        base_spiral = radius * np.array([np.cos(beta), np.sin(beta)])

        # move and scale fibonacci spiral to the previously calculated circle
        # use binary search to rescale until number of sample points
        # in the convex hull meets the condition

        return _binary_rescale(
            self._n_samples, _sample_generator(base_spiral, midpoint, ineqs), r
        )


class ArchimedeanSampler(Sampler):
    """A sampler that produces a number of approximately equidistant
    sample points on a moved and scaled version of the archimedean spiral
    (x * cos(x), x * sin(x)).

    The sample points all lie in the smallest enclosing circle
    of given data points.

    Inspired by
    https://agupubs.onlinelibrary.wiley.com/doi/full/10.1029/2007GC001581.

    Parameters
    ----------
    n_samples : positive int
        Amount of samples that will be produced by the sampler.

    Raises
    ------
    SamplerInitializationException
        If `n_samples` is non-positive.
    """

    def __init__(self, n_samples):
        if n_samples <= 0:
            raise SamplerInitializationException("`n_samples` is non-positive")

        self._n_samples = n_samples

    def sample(self, pts):
        """Produces samples according to the procedure described above.

        Parameters
        ----------
        pts : array_like of shape (n, 2)
            Points in whose convex hull the produced samples will lie.

        Returns
        -------
        samples : numpy.ndarray of shape (n_samples, 2)
            Samples produced by the procedure described above.
        """

        # calculate enclosing circle
        midpoint, r = _make_circle(pts)

        # compute convex hull and its volume
        ch = ConvexHull(pts)
        vol = ch.volume
        ineqs = ch.equations

        ub_n_samples = int(np.pi * r**2 * self._n_samples / vol) + 10

        # estimate upper bounds for spiral points needed
        # and simultaneously create arc values
        beta = [0]
        for _ in range(ub_n_samples):
            beta.append(beta[-1] + 8 / (beta[-1] + 2))

        beta = np.array(beta)
        base_spiral = beta * np.array([np.cos(beta), np.sin(beta)])

        return _binary_rescale(
            self._n_samples, _sample_generator(base_spiral, midpoint, ineqs), r
        )


def _binary_rescale(n_samples, generate_sample, start_value):
    """"""
    # performs binary search to create a sample
    # of size n_samples by calling generate_sample(t),
    # where t is the parameter modified
    lb = 0
    ub = None
    t = start_value
    sample = generate_sample(t)

    while len(sample) != n_samples:
        if len(sample) > n_samples:
            lb = t
        else:
            ub = t
        if ub is None:
            t *= 2
        else:
            t = (ub + lb) / 2
        sample = generate_sample(t)

    return sample


def _sample_generator(base_set, midpoint, ineqs):
    """"""

    # creates a function which generates samples as a scaled base_set
    # translated in midpoint and forfilling the
    # inequalities ineqs
    def generate_sample(t):
        spiral = (midpoint[:, None] + t * base_set).transpose()
        mask = np.all((ineqs[:, :2] @ spiral.T).T <= -ineqs[:, 2], axis=1)
        return spiral[mask]

    return generate_sample


def _make_circle(pts, eps=0.0001):
    """"""
    pts = pts.copy()
    np.random.shuffle(pts)
    k = []  # list of necessary boundary points
    circle = _small_circle(pts[k])
    i = 0  # index of currently examined point
    j = 0  # index of point to be included
    while i < len(pts):
        p = pts[i]

        if _is_in_circle(p, circle, eps):
            i += 1
            j = max(i, j)
            continue

        # p is not in circle
        if i == j:
            k = [i]
            i = 0
            circle = _small_circle(pts[k])
            continue

        # p is not in circle and len(k) < 3
        k = [t for t in k if t > i]
        k.append(i)
        circle = _small_circle(pts[k])
        i = 0

    return circle


def _is_in_circle(p, circle, eps):
    """"""
    mp, r = circle
    return np.linalg.norm(p - mp) < r + eps


def _small_circle(pts):
    """"""
    if len(pts) == 0:
        return np.zeros((2,)), 0
    if len(pts) == 1:
        return pts[0], 0
    if len(pts) == 2:
        p1, p2 = pts[0], pts[1]
        mp = 1 / 2 * (p1 + p2)
        r = 1 / 2 * np.linalg.norm(p1 - p2)
        return mp, r
    if len(pts) == 3:
        circle_m = -np.column_stack(pts).T
        circle_m = np.column_stack([np.ones(3), circle_m])
        circle_b = np.array([-np.linalg.norm(p) ** 2 for p in pts])
        # TODO: handling for degenerate case
        a, b, c = np.linalg.inv(circle_m) @ circle_b
        return np.array([b / 2, c / 2]), np.sqrt(b**2 / 4 + c**2 / 4 - a)
    raise ValueError(f"number of points should be <= 3 but is {pts}")



# pylint: disable=missing-module-docstring

import numpy as np


def scaled_norm(norm, scal_factors):
    scaled = np.asarray(scal_factors)

    def s_norm(vec):
        return norm(scaled * vec)

    return s_norm


def euclidean_norm(vec):
    return np.linalg.norm(vec, axis=1)


def scaled_euclidean_norm(vec):
    if vec.shape[1] == 2:
        norm_val = scaled_norm(euclidean_norm, [1 / 40, 1 / 360])(vec)
    elif vec.shape[1] == 3:
        norm_val = scaled_norm(euclidean_norm, [1 / 40, 1 / 360, 1 / 20])(vec)
    else:
        raise NotImplementedError(
            "scaled_euclidean_norm only supports 2 and 3 dimensional inputs"
        )
    return norm_val


def data_dict_to_numpy(data_dict, keys):
    """
    Method to transform a dictionary of lists of floats into a numpy array.

    Parameter
    ---------
    data_dict : dict,
        Dictionary to transform.

    keys : [str],
        Keys that indicate which lists of the data dictionary should be used
        to create the columns of the resulting array.

    Returns
    --------
    (n, d) array where 'n' is the length of a list in the data dictionary and
    'd' is 'len(keys)'.
    """
    return np.column_stack([data_dict[key] for key in keys])


def _safe_operation(operand, value):
    try:
        return operand(value)
    except (ValueError, TypeError, IndexError, KeyError, ZeroDivisionError):
        return None


class ComponentWithStatistics:
    """
    Interface class for pipelinecomponents enabling to save and handle run statistics.
    """

    def __init__(self):
        self._statistics = {}

    def set_statistics(self, **kwargs):
        """
        Sets the statistics dict corresponding to the keyword arguments.
        Supposed to be overwritten by inheriting classes with stronger requirements.
        """
        self._statistics = kwargs

    def get_latest_statistics(self):
        """
        Return
        ------
        statistics: dict
            The statistics of the latest run of the handler.
        """
        return self._statistics



"""
Classes used to model various geometric shapes centered around
the geometric origin.

Defines the `Neighbourhood` abstract base class that can be used
to create custom geometric shapes.

Subclasses of `Neighbourhood` can be used with the `TableExtension` and
the `PointcloudExtension` classes in the `hrosailing.pipeline` module.
"""


from abc import ABC, abstractmethod
from typing import Callable

import numpy as np

from ._utils import scaled_euclidean_norm


class NeighbourhoodInitializationException(Exception):
    """Exception raised if an error occurs during
    initialization of a `Neighbourhood`.
    """


class Neighbourhood(ABC):
    """Base class for all neighbourhood classes.


    Abstract Methods
    ----------------
    is_contained_in(self, pts)
    """

    @abstractmethod
    def is_contained_in(self, pts):
        """This method should be used, given certain points, to
        determine which of these points lie in the neighbourhood
        and which do not, by producing a boolean array of the same
        size as `pts`.

        Parameters
        ----------
        pts : array_like of shape (n, 2)
            Points that will be checked for membership.

        Returns
        -------
        mask : numpy.ndarray of shape (n, )
            Boolean array describing which of the input points
            are members of the neighbourhood.
        """


class Ball(Neighbourhood):
    """A class to describe a closed 2-dimensional ball
    centered around the origin, i.e. :math:`\\\\{x \\in R^2 : ||x|| <= r\\\\}`.

    Parameters
    ----------
    radius : positive int or float, optional
        The radius of the ball, i.e. :math:`r`.

        Defaults to `0.05`.

    norm : function or callable, optional
        The norm for which the ball is described, i.e. :math:`||.||`.

        If nothing is passed, it will default to a scaled version
        of :math:`||.||_2` in two dimensions.

    Raises
    ------
    NeighbourhoodInitializationException
        If `radius` is non-positive.
    """

    def __init__(
        self,
        radius=0.05,
        norm: Callable = scaled_euclidean_norm,
    ):
        if radius <= 0:
            raise NeighbourhoodInitializationException(
                "`radius` is non-positive"
            )

        self._norm = norm
        self._radius = radius

    def __repr__(self):
        return f"Ball(norm={self._norm.__name__}, radius={self._radius})"

    def is_contained_in(self, pts):
        """Checks given points for membership.

        Parameters
        ----------
        pts : array_like of shape (n, 2)
            Points that will be checked for membership.

        Returns
        -------
        mask : numpy.ndarray of shape (n, )
            Boolean array describing which of the input points
            are members of the neighbourhood.
        """
        pts = np.asarray(pts)
        try:
            return self._norm(pts) <= self._radius
        except ValueError:
            return np.asarray([self._norm(pt) <= self._radius for pt in pts])


class ScalingBall(Neighbourhood):
    """A class to represent a closed 2-dimensional ball
    centered around the origin, i.e. :math:`\\\\{x \\in R^2 : ||x|| <= r\\\\}`,
    where the radius :math:`r` will be dynamically determined, such that
    there is always a certain amount of given points contained
    in the ball.

    Parameters
    ----------
    min_pts : positive int
        The minimal amount of certain given points that should be
        contained in the scaling ball.

    norm : function or callable, optional
        The norm for which the scaling ball is described, i.e. :math:`||.||`.

        If nothing is passed, it will default to a scaled version
        of :math:`||.||_2`.

    Raises
    ------
    NeighbourhoodInitializationException
        If `min_pts` is non-positive.

    """

    def __init__(
        self,
        min_pts,
        norm: Callable = scaled_euclidean_norm,
    ):
        if min_pts <= 0:
            raise NeighbourhoodInitializationException(
                "`min_pts` is non-positive"
            )

        self._min_pts = min_pts
        self._norm = norm

        self._n_pts = None
        self._area = None
        self._radius = None

    def __repr__(self):
        return (
            f"ScalingBall(min_pts={self._min_pts}, norm={self._norm.__name__})"
        )

    def is_contained_in(self, pts):
        """Checks given points for membership, and scales
        ball so that at least `min_pts` points are contained in it.

        Parameters
        ----------
        pts : array_like of shape (n, 2)
            Points that will be checked for membership.

        Returns
        -------
        points_in_ball : boolean numpy.ndarray of shape (n, )
            Boolean array describing which of the input points
            are members of the neighbourhood.
        """
        pts = np.asarray(pts)

        dist = self._norm(pts)
        self._radius = sorted(dist)[self._min_pts - 1]

        return dist <= self._radius


class Ellipsoid(Neighbourhood):
    """A class to represent a closed d-dimensional ellipsoid
    centered around the origin, i.e. :math:`T(B)`, where :math:`T` is an invertible
    linear transformation, and :math:`B` is a closed d-dimensional ball,
    centered around the origin.

    It will be represented using the equivalent formulation:
    :math:`\\\\{x \\in R^2 : ||T^{-1} x|| \\leq r\\\\}`.

    Parameters
    ----------
    lin_trans : array_like of shape (2,2), optional
        The linear transformation which transforms the
        ball into the given ellipsoid, i.e. :math:`T`.

        If nothing is passed, it will default to the 2x2
        identity matrix, i.e. the ellipsoid will be a ball.

    norm : function or callable, optional
        The norm for which the ellipsoid is described, i.e. :math:`||.||`.

        If nothing is passed, it will default to a scaled
        version of :math:`||.||_2`.

    radius : positive int or float, optional
        The radius of the ellipsoid, i.e. :math:`r`.

        Defaults to `0.05`.


    Raises
    ------
    NeighbourhoodInitializationException
        If `radius` is non-positive.
     NeighbourhoodInitializationException
        If `lin_trans` is not a (2,2)-array or is not invertible.
    """

    def __init__(
        self,
        lin_trans=None,
        norm: Callable = scaled_euclidean_norm,
        radius=0.05,
    ):
        if lin_trans is None:
            lin_trans = np.eye(2)

        lin_trans = np.asarray_chkfinite(lin_trans)

        if lin_trans.shape != (2, 2):
            raise NeighbourhoodInitializationException(
                "`lin_trans` has incorrect shape"
            )

        if not np.linalg.det(lin_trans):
            raise NeighbourhoodInitializationException(
                "`lin_trans` is singular"
            )

        if radius <= 0:
            raise NeighbourhoodInitializationException(
                "`radius` is non-positive"
            )

        self._T = np.linalg.inv(lin_trans)
        self._norm = norm
        self._radius = radius

    def __repr__(self):
        return (
            f"Ellipsoid(lin_trans={self._T}, "
            f"norm={self._norm.__name__}, radius={self._radius})"
        )

    def is_contained_in(self, pts):
        """Checks given points for membership.

        Parameters
        ----------
        pts : array_like of shape (n, 2)
            Points that will be checked for membership.

        Returns
        -------
        mask : numpy.ndarray of shape (n, )
            Boolean array describing which of the input points
            are members of the neighbourhood.
        """
        pts = np.asarray(pts)

        pts = self._transform_ellipsoid_to_ball(pts)

        return self._norm(pts) <= self._radius

    def _transform_ellipsoid_to_ball(self, pts):
        return (self._T @ pts.T).T


class Cuboid(Neighbourhood):
    """A class to represent a d-dimensional closed cuboid, i.e.
    :math:`\\\\{x \\in R^2 : |x_i| \\leq b_i, i=1,2\\\\}`.

    Parameters
    ----------
    norm : function or callable, optional
        The 1-d norm used to measure the length of :math:`x_i`, i.e. :math:`|.|`.

        If nothing is passed, it will default to the absolute value :math:`|.|`.

    dimensions : subscriptable of length 2, optional
        The length of the sides of the cuboid, i.e. :math:`b_i`.

        If nothing is passed, it will default to `(0.05, 0.05)`.

    """

    def __init__(
        self,
        norm: Callable = np.abs,
        dimensions=(0.05, 0.05),
    ):
        self._norm = norm
        self._size = dimensions

    def __repr__(self):
        return f"Cuboid(norm={self._norm.__name__}, dimensions={self._size})"

    def is_contained_in(self, pts):
        """Checks given points for membership.

        Parameters
        ----------
        pts : array_like of shape (n, 2)
            Points that will be checked for membership.

        Returns
        -------
        mask : numpy.ndarray of shape (n, )
            Boolean array describing which of the input points
            are members of the neighbourhood.
        """
        pts = np.asarray(pts)

        mask = (
            np.ones((pts.shape[0],), dtype=bool)
            & (self._norm(pts[:, 0]) <= self._size[0])
            & (self._norm(pts[:, 1]) <= self._size[1])
        )
        return mask


class Polytope(Neighbourhood):
    """A class to represent a general 2-dimensional polytope, i.e. the
    convex hull :math:`P = conv(x_1, ..., x_n)` of some n points :math:`x_1 ,..., x_n`
    or equivalently the (bounded) intersection of m half spaces
    :math:`P = \\\\{x \\in R^2 : Ax \\leq b\\\\}`.

    Parameters
    ----------
    mat : array_like of shape (m, 2), optional
        Matrix to represent the normal vectors :math:`a_i` of the half
        spaces, i.e. :math:`A = (a_1, ... , a_m)^t`.

        If nothing is passed, it will default to :math:`(I_2, -I_2)^t`,
        where :math:`I_d` is the d-dimensional identity matrix.

    b : array_like of shape (m, ), optional
        Vector representing the :math:`b` of the half spaces, i.e.
        :math:`b = (b_1, ... , b_m)^t`.

        If nothing is passed, it will default to `(0.05,...,0.05)`.


    Raises
    ------
    NeighbourhoodInitializationException
        If `mat` and `b` are not of matching shape.


    Warning
    -------
    Does not check whether the polytope given by `mat` and `b` is a polytope,
    i.e. if :math:`P` is actually bounded.
    """

    def __init__(
        self, mat=np.row_stack((np.eye(2), -np.eye(2))), b=0.05 * np.ones(4)
    ):
        # NaN's or infinite values can't be handled
        mat = np.asarray_chkfinite(mat)
        b = np.asarray_chkfinite(b)

        if mat.ndim != 2 or mat.shape[1] != 2:
            raise NeighbourhoodInitializationException(
                "`mat` has incorrect shape"
            )

        if b.ndim != 1 or b.shape[0] != mat.shape[0]:
            raise NeighbourhoodInitializationException(
                "`b` has incorrect shape"
            )

        self._mat = mat
        self._b = b

    def __repr__(self):
        return f"Polytope(mat={self._mat}, b={self._b})"

    def is_contained_in(self, pts):
        """Checks given points for membership.

        Parameters
        ----------
        pts : array_like of shape (n, 2)
            Points that will be checked for membership.

        Returns
        -------
        mask : numpy.ndarray of shape (n, )
            Boolean array describing which of the input points
            are members of the neighbourhood.
        """
        pts = np.asarray(pts)

        mask = np.ones((pts.shape[0],), dtype=bool)
        for ineq, bound in zip(self._mat, self._b):
            mask = mask & (ineq @ pts.T <= bound)
        return mask



"""
Contains the Base class for the `Imputator` pipeline-component that can be used
to create custom imputators.
"""

from abc import ABC, abstractmethod
from datetime import timedelta

from hrosailing.pipelinecomponents._utils import ComponentWithStatistics


class Imputator(ComponentWithStatistics, ABC):
    """Base class for all imputator classes.


    Abstract Methods
    ----------------
    impute(self, data)
    """

    @abstractmethod
    def impute(self, data):
        """
        This method should be used, given data that possibly contains `None`
        values, to create data without any `None` values.

        Parameters
        ----------
        data : Data
            Data to be imputed.
        """

    def set_statistics(
        self, n_removed_cols, n_removed_rows, n_filled, data_dict
    ):
        super().set_statistics(
            n_removed_cols=n_removed_cols,
            n_removed_rows=n_removed_rows,
            n_filled_fields=n_filled,
            n_rows=data_dict.n_rows,
            n_cols=data_dict.n_cols,
        )


class RemoveOnlyImputator(Imputator):
    """
    Imputator that removes all entries containing None values.
    """

    def impute(self, data):
        """
        Removes all entries containing None values.

        See also
        --------
        `Imputator.impute`
        """
        n_rows = data.n_rows
        remove_rows = [
            idx
            for idx, row in enumerate(data.rows(return_type=list))
            if None in row
        ]
        data.delete(remove_rows)

        self.set_statistics(0, n_rows - len(remove_rows), 0, data)

        return data


class FillLocalImputator(Imputator):
    """
    An `Imputator` which assumes that the data has been stored chronologically
    and contains the field 'datetime'.
    Fills missing data by:
    - deleting columns that only contain `None` values,
    - deleting rows between two data-points that are far apart in time,
    - affine interpolation of datetime-stamps between two data-points which
        are not far apart in time,
    - filling data before and after a not-`None` value which are not too far apart
        in time according to certain functions (`fill_before` and `fill_after`),
    - filling data between a pair of not-`None` values which are not too far apart
        in time according to a certain function (`fill_between`).

    Parameters
    ----------
    fill_before : (str, object, float) -> object, optional
        The function which will be used to fill `None` fields before a not-`None`
        field.

        First argument is the attribute name, second is the value of the
        not-`None` field mentioned before, third argument is the relative
        position (in time) between the earliest applicable data and the
        mentioned not-`None` data. Returns the value that replaces `None`.

        Defaults to `lambda name, right, mu: right`.

    fill_between : (str, object, object, float) -> object
        The function which will be used to fill `None` fields between
        two not-`None` fields.

        First argument is the attribute `name`, second and third are the values
        of the not-`None` fields mentioned before in chronological order,
        last argument is the relative position (in time) between
        the two mentioned not-`None` data points.
        Returns the value to be filled.

        Defaults to `lambda name, left, right, mu: left`.

    fill_after : (str, object, float) -> object, optional
        The function which will be used to fill `None` fields after a not-`None`
        field.

        First argument is the attribute `name`, second is the value of the
        not-`None` field mentioned before, third argument is the relative
        position (in time) between the mentioned not-`None` data and the
        latest applicable data. Returns the value to be filled.

        Defaults to `lambda name, left, mu: left`.

    max_time_diff : datetime.timedelta, optional
        Two data points are treated as 'close in time' if their time difference
        is smaller than `max_time_diff`.

        Defaults to `2 minutes`.
    """

    def __init__(
        self,
        fill_before=lambda name, right, mu: right,
        fill_between=lambda name, left, right, mu: left,
        fill_after=lambda name, left, mu: left,
        max_time_diff=timedelta(minutes=2),
    ):
        super().__init__()
        self._fill_before = lambda name, left, right, mu: fill_before(
            name, right, mu
        )
        self._fill_between = fill_between
        self._fill_after = lambda name, left, right, mu: fill_after(
            name, left, mu
        )
        self._max_time_diff = max_time_diff
        self._n_filled = 0

    def impute(self, data):
        """
        Creates a dictionary that does not contain `None` values by the procedure
        described above.

        Parameters
        ----------
        data : Data
            The `Data` object to be imputed.

        Returns
        -------
        data_dict : Data
            `data_dict` is the resulting `Data` object containing no `None` values.
        """
        self._n_filled = 0
        n_removed_cols = data.n_cols
        data.strip("cols")
        n_removed_cols -= data.n_cols

        data = self._interpolate_datetime(data)

        data, n_removed_rows = self._remove_rows(data)

        data = self._interpolate_other(data)

        # remove rows which still have None values
        remove_rows = [
            i
            for i, _ in enumerate(data["datetime"])
            if any(data[key][i] is None for key in data.keys())
        ]

        data.delete(remove_rows)
        n_removed_rows += len(remove_rows)

        self.set_statistics(
            n_removed_cols, n_removed_rows, self._n_filled, data
        )

        return data

    def _fill_range(
        self, data_dict, datetime, key, start_idx, end_idx, fill_fun
    ):
        left = data_dict[key][start_idx]
        right = data_dict[key][end_idx]
        for i in range(start_idx + 1, end_idx):
            duration = datetime[end_idx] - datetime[start_idx]
            try:
                mu = (datetime[i] - datetime[start_idx]) / duration
            except ZeroDivisionError:
                mu = 0
            data_dict[key][i] = fill_fun(key, left, right, mu)
            self._n_filled += 1

    def _interpolate_datetime(self, data):
        last_dt, last_i = None, None
        for i, dt in enumerate(data["datetime"]):
            if dt is None:
                continue
            if last_dt is None:
                pass
            elif abs(dt - last_dt) > self._max_time_diff:
                continue
            else:
                # linear approximation of time in between
                for j in range(last_i + 1, i):
                    mu = (j - last_i) / (i - last_i)
                    data["datetime"][j] = last_dt + mu * (dt - last_dt)
            last_dt, last_i = dt, i

        return data

    def _remove_rows(self, data):
        remove_rows = [
            i for i, dt in enumerate(data["datetime"]) if dt is None
        ]
        data.delete(remove_rows)

        n_removed_rows = len(remove_rows)

        return data, n_removed_rows

    def _interpolate_other(self, data):
        # indices of not None values
        idx_dict = {
            key: [i for i, data in enumerate(data[key]) if data is not None]
            for key in data.keys()
        }

        datetime = data["datetime"]

        for key, idx in idx_dict.items():
            # fill every entry before the first not-None entry according to the
            # "fill before" function
            if key == "datetime":
                continue

            try:
                start_idx = self._aplly_fill_before(datetime, idx, key, data)
            except ValueError:
                continue

            self._apply_fill_between(key, idx, data, datetime)

            self._apply_fill_after(key, idx, data, datetime, start_idx)

        return data

    def _aplly_fill_before(self, datetime, idx, key, data):
        if idx[0] > 0:
            start_idx = min(
                i
                for i in range(idx[0])
                if datetime[idx[0]] - datetime[i] < self._max_time_diff
            )  # first idx in time interval

            self._fill_range(
                data,
                datetime,
                key,
                start_idx,
                idx[0],
                self._fill_before,
            )

            return start_idx
        return None

    def _apply_fill_between(self, key, idx, data, datetime):
        for idx1, idx2 in zip(idx, idx[1:]):
            timediff = datetime[idx2] - datetime[idx1]
            if timediff < self._max_time_diff:
                # fill data according to fill_between function
                self._fill_range(
                    data,
                    datetime,
                    key,
                    idx1,
                    idx2,
                    self._fill_between,
                )
            else:
                # fill data according to fill_before and fill_after
                near_points = [
                    i
                    for i in range(idx1 + 1, idx2)
                    if datetime[i] - datetime[idx1] < self._max_time_diff
                ]
                if len(near_points) > 0:
                    last_idx_right = max(near_points)
                    self._fill_range(
                        data,
                        datetime,
                        key,
                        idx1,
                        last_idx_right,
                        self._fill_after,
                    )
                near_points = [
                    i
                    for i in range(idx1 + 1, idx2)
                    if datetime[idx2] - datetime[i] < self._max_time_diff
                ]
                if len(near_points) > 0:
                    first_idx_left = min(near_points)
                    self._fill_range(
                        data,
                        datetime,
                        key,
                        first_idx_left,
                        idx2,
                        self._fill_before,
                    )

    def _apply_fill_after(self, key, idx, data, datetime, start_idx):
        near_points = [
            i
            for i in range(idx[0])
            if datetime[idx[0]] - datetime[i] < self._max_time_diff
        ]
        if len(near_points) > 0:
            end_idx = min(near_points)  # first idx in time interval
            self._fill_range(
                data,
                datetime,
                key,
                start_idx,
                end_idx,
                self._fill_after,
            )



"""Functions to convert wind from apparent to true and vice versa."""


import enum

import numpy as np


class WindConversionException(Exception):
    """Exception raised if an error occurs during wind conversion"""


class _Wind(enum.Enum):
    TO_TRUE = -1
    TO_APPARENT = 1

    def convert_wind(self, wind):
        wind = np.asarray_chkfinite(wind)

        if wind.dtype == object:
            raise WindConversionException("`wind` is not array_like")
        if wind.ndim != 2 or wind.shape[1] != 3:
            raise WindConversionException("`wind` has incorrect shape")

        ws, wa, bsp = np.hsplit(wind, 3)
        if np.any((ws < 0)):
            raise WindConversionException("`wind` has negative wind speeds")

        wa %= 360  # normalize wind angles
        wa_above_180 = wa > 180
        wa = np.deg2rad(wa)

        converted_ws = self._convert_wind_speed(ws, wa, bsp)
        converted_wa = self._convert_wind_angle(
            converted_ws, ws, wa, bsp, wa_above_180
        )

        return np.column_stack((converted_ws, converted_wa, bsp))

    def _convert_wind_speed(self, ws, wa, bsp):
        return np.sqrt(
            np.square(ws)
            + np.square(bsp)
            + 2 * self.value * ws * bsp * np.cos(wa)
        )

    def _convert_wind_angle(self, converted_ws, ws, wa, bsp, wa_above_180):
        temp = (ws * np.cos(wa) + self.value * bsp) / converted_ws

        # account for floating point errors
        temp[temp > 1] = 1
        temp[temp < -1] = -1

        converted_wa = np.arccos(temp)
        converted_wa[wa_above_180] = 360 - np.rad2deg(
            converted_wa[wa_above_180]
        )
        converted_wa[~wa_above_180] = np.rad2deg(converted_wa[~wa_above_180])

        return converted_wa


def convert_apparent_wind_to_true(apparent_wind):
    """Convert apparent wind to true wind.

    Parameters
    ----------
    apparent_wind : array_like of shape (n, 3)
        Wind data given as a sequence of points consisting of wind speed,
        wind angle and boat speed, where the wind speed and wind angle are
        measured as apparent wind.

    Returns
    -------
    true_wind : numpy.ndarray of shape (n, 3)
        Array containing the same data as `apparent_wind`, but the wind speed
        and wind angle now measured as true wind.

    Raises
    ------
    WindConversionException
    """
    return _Wind.TO_TRUE.convert_wind(apparent_wind)


def convert_true_wind_to_apparent(true_wind):
    """Convert true wind to apparent wind.

    Parameters
    ----------
    true_wind : array_like of shape (n, 3)
        Wind data given as a sequence of points consisting of wind speed,
        wind angle and boat speed, where the wind speed and wind angle are
        measured as true wind.

    Returns
    -------
    apparent_wind : numpy.ndarray of shape (n, 3)
        Array containing the same data as `true_wind`, but the wind speed
        and wind angle now measured as apparent wind.

    Raises
    ------
    WindConversionException
    """
    return _Wind.TO_APPARENT.convert_wind(true_wind)
